from .managers.BroadcastManager import Broadcast
from .managers.CacheManager import Cache
from .managers.MailManager import Mail
from .managers.QueueManager import Queue
from .managers.SessionManager import Session
from .managers.UploadManager import Upload
from .environment import env
from .__version__ import (__title__, __description__, __url__,
                          __version__, __author__, __author_email__,
                          __licence__)



class NoneType:
    def __getattr__(self, attr):
        return None

    def __call__(self, *args, **kwargs):
        return None

    def __eq__(self, other):
        return other is None


class Optional:

    def __init__(self, obj):
        self.obj = obj

    def __getattr__(self, attr):
        if hasattr(self.obj, attr):
            return getattr(self.obj, attr)
        return NoneType()

    def __call__(self, *args, **kwargs):
        return NoneType()

    def instance(self):
        return self.obj



"""View Helper Module."""

from jinja2 import Markup


def set_request_method(method_type):
    """Return an input string for use in a view to change the request method of a form.

    Arguments:
        method_type {string} -- Can be options like GET, POST, PUT, PATCH, DELETE

    Returns:
        string -- An input string.
    """
    return Markup("<input type='hidden' name='__method' value='{0}'>".format(method_type))


def back(location=None):
    """Return an input element for use in telling Masonite which route to redirect back to.

    Arguments:
        location {string} -- The route to redirect back to.

    Returns:
        string -- An input string.
    """
    if location is None:
        from wsgi import container
        location = container.make('Request').path

    return Markup("<input type='hidden' name='__back' value='{0}'>".format(location))


def hidden(value, name='hidden-input'):
    return Markup("<input type='hidden' name='{}' value='{}'>".format(name, value))


def old(session_key, default=''):
    """Return the old value submitted by forms validated with Valitators.

    Arguments:
        session_key {string} -- The key flashed to session.

    Returns:
        string -- An input string.
    """

    from wsgi import container
    session_container = container.make('Session')

    if session_container.has(session_key):
        return session_container.get(session_key)
    return default



"""Mail Manager Module."""

from ..contracts import MailManagerContract
from .Manager import Manager
from ..helpers import config


class MailManager(Manager, MailManagerContract):
    """Manages all mail drivers.

    Arguments:
        Manager {from .managers.Manager} -- The base Manager class.
    """

    config = 'mail'
    driver_prefix = 'Mail'

    def helper(self):
        """Helper Method to work with returning the driver from the MailManager.

        Returns:
            Mail Driver
        """
        return self.driver(config('mail.driver'))


class Mail:
    """Dummy class that will be used to swap out the manager in the container."""

    pass



"""Password Helper Module."""

import bcrypt


def password(password):
    """Bcrypt a string.

    Useful for storing passwords in a database.

    Arguments:
        password {string} -- A string like a users plain text password to be bcrypted.

    Returns:
        string -- The encrypted string.
    """
    return bytes(bcrypt.hashpw(
        bytes(password, 'utf-8'), bcrypt.gensalt()
    )).decode('utf-8')



"""Base Queue Module."""


class Queueable:
    """Makes classes Queueable."""

    run_again_on_fail = True
    run_times = 3

    def handle(self):
        """Put the queue logic in this handle method."""
        pass

    def dispatch(self):
        """Responsible for dispatching the job to the Queue service.

        Returns:
            self.handle
        """
        return self.handle

    def should_run(self, job):
        return True



"""CSRF Middleware."""

from jinja2 import Markup

from ..auth import Csrf
from ..exceptions import InvalidCSRFToken
from ..request import Request
from ..view import View


class CsrfMiddleware:
    """Verify CSRF Token Middleware."""

    exempt = ['/']
    every_request = True
    token_length = 30

    def __init__(self, request: Request, csrf: Csrf, view: View):
        """Initialize the CSRF Middleware

        Arguments:
            request {masonite.request.Request} -- The normal Masonite request class.
            csrf {masonite.auth.Csrf} -- CSRF auth class.
            view {masonite.view.View} -- The normal Masonite view class.
        """

        self.request = request
        self.csrf = csrf
        self.view = view

    def before(self):
        """Execute this method before the controller."""

        token = self.verify_token()

        self.view.share({
            'csrf_field': Markup("<input type='hidden' name='__token' value='{0}' />".format(token)),
            'csrf_token': token
        })

    def after(self):
        pass

    def in_exempt(self):
        """Determine if the request has a URI that should pass through CSRF verification.

        Returns:
            bool
        """
        for route in self.exempt:
            if self.request.contains(route):
                return True

        return False

    def generate_token(self):
        """Generate a token that will be used for CSRF protection

        Returns:
            string -- A random string based on the length given
        """

        return self.csrf.generate_csrf_token(self.token_length)

    def verify_token(self):
        """Verify if csrf token in post is valid.

        Raises:
            InvalidCSRFToken -- Thrown if the CSRF tokens do not match.

        Returns:
            string -- Returns a new token or the current token.
        """

        if self.request.is_post() and not self.in_exempt():
            token = self.request.input('__token')
            if not self.csrf.verify_csrf_token(token):
                raise InvalidCSRFToken("Invalid CSRF token.")
            return token
        else:
            if not self.every_request and self.request.get_cookie('csrf_token', decrypt=False):
                return self.request.get_cookie('csrf_token', decrypt=False)
            else:
                return self.generate_token()



"""Base queue driver."""

import pickle

import pendulum

from ...drivers import BaseDriver
from ...helpers import HasColoredCommands


class BaseQueueDriver(BaseDriver, HasColoredCommands):

    def add_to_failed_queue_table(self, payload, driver='amqp'):
        from config.database import DB as schema
        from config import queue
        if 'amqp' in queue.DRIVERS:
            listening_channel = queue.DRIVERS['amqp']['channel']
        else:
            listening_channel = 'default'

        if schema.get_schema_builder().has_table('failed_jobs'):
            schema.table('failed_jobs').insert({
                'driver': driver,
                'channel': listening_channel,
                'payload': pickle.dumps(payload),
                'failed_at': pendulum.now()
            })

    def run_failed_jobs(self):
        from config.database import DB as schema
        try:
            self.success('Attempting to send failed jobs back to the queue ...')
            for job in schema.table('failed_jobs').get():
                payload = pickle.loads(job.payload)
                schema.table('failed_jobs').where('payload', job.payload).delete()
                self.push(payload['obj'], args=payload['args'], callback=payload['callback'])
        except Exception:
            self.danger('Could not get the failed_jobs table')

    def push(self, *objects, args=(), callback='handle', ran=1, channel=None):
        raise NotImplementedError

    def connect(self):
        return self

    def consume(self, channel, fair=False):
        raise NotImplementedError('The {} driver does not implement consume'.format(self.__class__.__name__))

    def work(self):
        raise NotImplementedError('The {} driver does not implement work'.format(self.__class__.__name__))



from abc import ABC, abstractmethod


class QueueContract(ABC):

    @abstractmethod
    def push(self, *objects, args=(), callback='handle', ran=1, channel=None):
        pass

    @abstractmethod
    def connect(self):
        pass

    @abstractmethod
    def consume(self, channel, fair=False):
        pass

    @abstractmethod
    def work(self):
        pass

    @abstractmethod
    def run_failed_jobs(self):
        pass

    @abstractmethod
    def add_to_failed_queue_table(self):
        pass



import subprocess

from ..helpers import config, HasColoredCommands
from ..packages import add_venv_site_packages
from orator.migrations import DatabaseMigrationRepository, Migrator


class Migrations(HasColoredCommands):

    def __init__(self, connection=None):
        self._ran = []
        self._notes = []
        from config import database

        self.repository = DatabaseMigrationRepository(database.DB, 'migrations')
        self.migrator = Migrator(self.repository, database.DB)
        if not connection or connection == 'default':
            connection = database.DATABASES['default']
        self.migrator.set_connection(connection)
        if not self.repository.repository_exists():
            self.repository.create_repository()

        from wsgi import container

        self.migration_directories = ['databases/migrations']
        for key, value in container.providers.items():
            if isinstance(key, str) and 'MigrationDirectory' in key:
                self.migration_directories.append(value)

        try:
            add_venv_site_packages()
        except ImportError:
            self.comment(
                'This command must be ran inside of the root of a Masonite project directory')

    def run(self):
        for directory in self.migration_directories:
            try:
                if len(self.migration_directories) > 1:
                    self.info('Migrating: {}'.format(directory))
                self.migrator.run(directory)
                self._ran.append(self.repository.get_ran())
                self._notes = self.migrator._notes
            except Exception as e:
                self.danger(str(e))

        return self

    def rollback(self):
        for directory in self.migration_directories:
            try:
                if len(self.migration_directories) > 1:
                    self.info('Migrating: {}'.format(directory))
                self.migrator.rollback(directory)
                self._ran.append(self.repository.get_ran())
                self._notes = self.migrator._notes
            except Exception as e:
                self.danger(str(e))

        return self

    def refresh(self):
        self.run()
        self.rollback()

    def reset(self):
        for directory in self.migration_directories:
            try:
                if len(self.migration_directories) > 1:
                    self.info('Migrating: {}'.format(directory))
                self.migrator.reset(directory)
                self._ran.append(self.repository.get_ran())
                self._notes = self.migrator._notes
            except Exception as e:
                self.danger(str(e))

        return self

    def ran(self):
        return self._ran


def has_unmigrated_migrations():
    if not config('application.debug'):
        return False

    from wsgi import container
    from config.database import DB
    try:
        DB.connection()
    except Exception:
        return False

    migration_directory = ['databases/migrations']
    for key, value in container.providers.items():
        if isinstance(key, str) and 'MigrationDirectory' in key:
            migration_directory.append(value)

    for directory in migration_directory:
        try:
            output = bytes(subprocess.check_output(
                ['orator', 'migrate:status', '-c',
                    'config/database.py', '-p', directory]
            )).decode('utf-8')

            if 'No' in output:
                return True
        except Exception:
            pass

    return False



"""Module for using the Ably websocket driver."""

from ...contracts import BroadcastContract
from ...drivers import BaseDriver
from ...exceptions import DriverLibraryNotFound
from ...app import App


class BroadcastAblyDriver(BroadcastContract, BaseDriver):
    """Class for the Ably Driver."""

    def __init__(self, app: App):
        """Ably driver constructor.

        Arguments:
            BroadcastConfig {config.broadcast} -- Broadcast configuration setting.
        """
        self.config = app.make('BroadcastConfig')
        self.ssl_message = True

    def ssl(self, boolean):
        """Set whether to send data with SSL enabled.

        Arguments:
            boolean {bool} -- Boolean on whether to set SSL.

        Returns:
            self
        """
        self.ssl_message = boolean
        return self

    def channel(self, channels, message, event='base-event'):
        """Specify which channel(s) you want to send information to.

        Arguments:
            channels {string|list} -- Can be a string for the channel or a list of strings for the channels.
            message {string} -- The message you want to send to the channel(s)

        Keyword Arguments:
            event {string} -- The event you want broadcasted along with your data. (default: {'base-event'})

        Raises:
            DriverLibraryNotFound -- Thrown when ably is not installed.

        Returns:
            string -- Returns the message sent.
        """
        try:
            from ably import AblyRest
        except ImportError:
            raise DriverLibraryNotFound(
                'Could not find the "ably" library. Please pip install this library running "pip install ably"')

        client = AblyRest('{0}'.format(
            self.config.DRIVERS['ably']['secret']
        ))

        if isinstance(channels, list):
            for channel in channels:
                ably_channel = client.channels.get(channel)
                ably_channel.publish(event, message)
        else:
            channel = client.channels.get(channels)
            channel.publish(event, message)

        return message



"""A RedirectionProvider Service Provider."""


from ..drivers import QueueAsyncDriver, QueueAmqpDriver
from ..managers import QueueManager
from ..provider import ServiceProvider
from .. import Queue
from ..helpers import config


class QueueProvider(ServiceProvider):

    wsgi = False

    def register(self):
        self.app.bind('QueueAsyncDriver', QueueAsyncDriver)
        self.app.bind('QueueAmqpDriver', QueueAmqpDriver)
        self.app.bind('QueueManager', QueueManager)

    def boot(self, queue: QueueManager):
        self.app.bind('Queue', queue.driver(config('queue.driver')))
        self.app.swap(Queue, queue.driver(config('queue.driver')))



from abc import ABC as Contract, abstractmethod


class AuthContract(Contract):

    @abstractmethod
    def user(self):
        pass

    @abstractmethod
    def save(self):
        pass

    @abstractmethod
    def delete(self):
        pass



from ..request import Request
from ..testsuite import TestSuite, generate_wsgi
import json
from ..helpers import Dot


class MockRoute:

    def __init__(self, route, container, wsgi=None):
        self.route = route
        self.container = container
        self.wsgi = wsgi

    def isNamed(self, name):
        return self.route.named_route == name

    def hasMiddleware(self, *middleware):
        return all(elem in self.route.list_middleware for elem in middleware)

    def hasController(self, controller):
        return self.route.controller == controller

    def contains(self, value):
        return value in self.container.make('Response')

    def assertContains(self, value):
        assert value in self.container.make('Response'), "Response does not contain {}".format(value)
        return self

    def assertNotFound(self):
        return self.assertIsStatus(404)

    def ok(self):
        return '200 OK' in self.container.make('Request').get_status_code()

    def canView(self):
        wsgi = generate_wsgi()
        wsgi['PATH_INFO'] = self.route.route_url
        wsgi['RAW_URI'] = self.route.route_url
        self.container = self._run_container(wsgi).container

        return self.container.make('Request').get_status_code() == '200 OK'

    def hasJson(self, key, value=''):
        response = json.loads(self.container.make('Response'))
        if isinstance(key, dict):
            for item_key, key_value in key.items():
                if not Dot().dot(item_key, response, False) == key_value:
                    return False
            return True
        return Dot().dot(key, response, False)

    def assertHasJson(self, key, value):
        response = json.loads(self.container.make('Response'))
        if isinstance(key, dict):
            for item_key, key_value in key.items():
                assert Dot().dot(item_key, response, False) == key_value
        else:
            assert Dot().dot(key, response, False) == value, "Key '{}' with the value of '{}' could not find a match in {}".format(key, value, response)
        return self

    def assertJsonContains(self, key, value):
        response = json.loads(self.container.make('Response'))
        if not isinstance(response, list):
            raise ValueError("This method can only be used if the response is a list of elements.")

        found = False
        for element in response:
            if Dot().dot(key, element, False):
                assert Dot().dot(key, element, False)
                found = True

        if not found:
            raise AssertionError("Could not find a key of: {} that had the value of {}".format(key, value))
        return self

    def count(self, amount):
        return len(json.loads(self.container.make('Response'))) == amount

    def assertCount(self, amount):
        response_amount = len(json.loads(self.container.make('Response')))
        assert response_amount == amount, 'Response has an count of {}. Asserted {}'.format(response_amount, amount)
        return self

    def amount(self, amount):
        return self.count(amount)

    def hasAmount(self, key, amount):
        response = json.loads(self.container.make('Response'))
        try:
            return len(response[key]) == amount
        except TypeError:
            raise TypeError("The json response key of: {} is not iterable but has the value of {}".format(key, response[key]))

    def assertHasAmount(self, key, amount):
        response = json.loads(self.container.make('Response'))
        try:
            assert len(response[key]) == amount, '{} is not equal to {}'.format(len(response[key]), amount)
        except TypeError:
            raise TypeError("The json response key of: {} is not iterable but has the value of {}".format(key, response[key]))

        return self

    def assertNotHasAmount(self, key, amount):
        response = json.loads(self.container.make('Response'))
        try:
            assert not len(response[key]) == amount, '{} is equal to {} but should not be'.format(len(response[key]), amount)
        except TypeError:
            raise TypeError("The json response key of: {} is not iterable but has the value of {}".format(key, response[key]))

        return self

    def user(self, obj):
        self._user = obj
        self.container.on_resolve(Request, self._bind_user_to_request)
        return self

    def isPost(self):
        return 'POST' in self.route.method_type

    def isGet(self):
        return 'GET' in self.route.method_type

    def isPut(self):
        return 'PUT' in self.route.method_type

    def isPatch(self):
        return 'PATCH' in self.route.method_type

    def isDelete(self):
        return 'DELETE' in self.route.method_type

    def on_bind(self, obj, method):
        self.container.on_bind(obj, method)
        return self

    def hasSession(self, key):
        wsgi = generate_wsgi()
        wsgi['PATH_INFO'] = self.route.route_url
        wsgi['RAW_URI'] = self.route.route_url
        self.container = self._run_container(wsgi).container
        return self.container.make('Session').has(key)

    def assertParameterIs(self, key, value):
        request = self.container.make('Request')
        if key not in request.url_params:
            raise AssertionError("Request class does not have the '{}' url parameter".format(key))

        if request.param(key) != value:
            raise AssertionError('parameter {} is equal to {} of type {}, not {} of type {}'.format(key, request.param(key), type(request.param(key)), value, type(value)))

    def assertIsStatus(self, status):
        request = self.container.make('Request')
        assert request.is_status(status), AssertionError("{} is not equal to {}".format(request.get_status_code(), status))
        if not request.is_status(status):
            raise AssertionError("{} is not equal to {}".format(request.get_status_code(), status))

        return self

    def assertHasHeader(self, key):
        pass

    def assertHeaderIs(self, key, value):
        request = self.container.make('Request')
        assert str(request.header(key)) == str(value), AssertionError("{} is not equal to {}".format(request.header(key), value))

        return self

    def assertPathIs(self, url):
        path = self.container.make('Request').path
        assert path == url, "Asserting the path is '{}' but it is '{}'".format(url, path)
        return True

    def session(self, key):
        wsgi = generate_wsgi()
        wsgi['PATH_INFO'] = self.route.route_url
        wsgi['RAW_URI'] = self.route.route_url
        self.container = self._run_container(wsgi).container
        return self.container.make('Session').get(key)

    def on_make(self, obj, method):
        self.container.on_make(obj, method)
        return self

    def on_resolve(self, obj, method):
        self.container.on_resolve(obj, method)
        return self

    def _run_container(self, wsgi):
        return TestSuite().create_container(wsgi, container=self.container)

    def _bind_user_to_request(self, request, container):
        request.set_user(self._user)
        return self

    def headerIs(self, key, value):
        request = self.container.make('Request')
        assertion = request.header(key) == value
        if not assertion:
            raise AssertionError('header {} does not equal {}'.format(request.header(key), value))
        return assertion

    def parameterIs(self, key, value):
        request = self.container.make('Request')
        assertion = request.param(key) == value
        if not assertion:
            raise AssertionError('parameter {} is equal to {} of type {}, not {} of type {}'.format(key, request.param(key), type(request.param(key)), value, type(value)))
        return assertion

    @property
    def request(self):
        return self.container.make('Request')

    @property
    def response(self):
        return self.container.make('Response')

    def asDictionary(self):
        try:
            return json.loads(self.response)
        except ValueError:
            raise ValueError("The response was not json serializable")



"""New Authentication System Command."""
import os
import shutil

from cleo import Command


class AuthCommand(Command):
    """
    Creates an authentication system.

    auth
    """

    def handle(self):
        self.info('Scaffolding Application ...')
        module_path = os.path.dirname(os.path.realpath(__file__))

        with open('routes/web.py', 'a') as f:
            # add all the routes
            f.write('\nROUTES = ROUTES + [\n    ')
            f.write("Get().route('/login', 'auth.LoginController@show').name('login'),\n    ")
            f.write("Get().route('/logout', 'auth.LoginController@logout').name('logout'),\n    ")
            f.write("Post().route('/login', 'auth.LoginController@store'),\n    ")
            f.write("Get().route('/register', 'auth.RegisterController@show').name('register'),\n    ")
            f.write("Post().route('/register', 'auth.RegisterController@store'),\n    ")
            f.write("Get().route('/home', 'auth.HomeController@show').name('home'),\n    ")
            f.write("Get().route('/email/verify', 'auth.ConfirmController@verify_show').name('verify'),\n    ")
            f.write("Get().route('/email/verify/@id:signed', 'auth.ConfirmController@confirm_email'),\n    ")
            f.write("Get().route('/email/verify/@id:signed', 'auth.ConfirmController@confirm_email'),\n    ")
            f.write("Get().route('/password', 'auth.PasswordController@forget').name('forgot.password'),\n    ")
            f.write("Post().route('/password', 'auth.PasswordController@send'),\n    ")
            f.write("Get().route('/password/@token/reset', 'auth.PasswordController@reset').name('password.reset'),\n    ")
            f.write("Post().route('/password/@token/reset', 'auth.PasswordController@update'),\n")
            f.write(']\n')

        # move controllers
        shutil.copyfile(module_path + "/../snippets/auth/controllers/LoginController.py",
                        os.getcwd() + "/app/http/controllers/auth/LoginController.py")
        shutil.copyfile(module_path + "/../snippets/auth/controllers/RegisterController.py",
                        os.getcwd() + "/app/http/controllers/auth/RegisterController.py")
        shutil.copyfile(module_path + "/../snippets/auth/controllers/HomeController.py",
                        os.getcwd() + "/app/http/controllers/auth/HomeController.py")
        shutil.copyfile(module_path + "/../snippets/auth/controllers/ConfirmController.py",
                        os.getcwd() + "/app/http/controllers/auth/ConfirmController.py")
        shutil.copyfile(module_path + "/../snippets/auth/controllers/PasswordController.py",
                        os.getcwd() + "/app/http/controllers/auth/PasswordController.py")
        # move templates
        shutil.copytree(module_path + "/../snippets/auth/templates/auth",
                        os.getcwd() + "/resources/templates/auth")

        self.info('Project Scaffolded. You now have 5 new controllers, 7 new templates and 9 new routes')



import os
import shutil


def make_directory(directory):
    if not os.path.isfile(directory):
        if not os.path.exists(os.path.dirname(directory)):
            # Create the path to the model if it does not exist
            os.makedirs(os.path.dirname(directory))

        return True

    return False


def copy_migration(directory_file, to='databases/migrations'):
    import datetime
    base_path = os.path.join(os.path.dirname(os.path.realpath(__file__)), '../../')

    file_path = os.path.join(base_path, directory_file)
    to_location = os.path.join(os.getcwd(), to, datetime.datetime.utcnow().strftime("%Y_%m_%d_%H%M%S") + '_' + os.path.basename(directory_file))
    shutil.copyfile(file_path, to_location)

    print('\033[92m {} has been created \033[0m'.format(to_location))



"""Migrate Refresh Command."""

from cleo import Command


class MigrateRefreshCommand(Command):
    """
    Migrate refresh.

    migrate:refresh
        {--s|seed : Seed the database}
        {--c|connection=default : The connection you want to run migrations on}
    """

    def handle(self):
        self.call('migrate:reset', [
            ('--connection', self.option('connection'))
        ])
        self.call('migrate', [
            ('--connection', self.option('connection'))
        ])

        if self.option('seed'):
            self.call('seed:run')



"""A RedirectionProvider Service Provider."""

from ..drivers import BroadcastAblyDriver, BroadcastPusherDriver
from ..managers import BroadcastManager
from ..provider import ServiceProvider
from .. import Broadcast
from ..helpers import config


class BroadcastProvider(ServiceProvider):

    wsgi = False

    def register(self):
        self.app.bind('BroadcastPusherDriver', BroadcastPusherDriver)
        self.app.bind('BroadcastAblyDriver', BroadcastAblyDriver)
        self.app.bind('BroadcastManager', BroadcastManager(self.app))

    def boot(self):
        self.app.bind('Broadcast', self.app.make('BroadcastManager').driver(config('broadcast.driver')))
        self.app.swap(Broadcast, self.app.make('BroadcastManager').driver(config('broadcast.driver')))



"""CORS Middleware."""

from ..helpers import config
from ..request import Request


class CorsMiddleware:
    """Appends CORS headers to HTTP response.

    Put any CORS middleware you need as a CORS dictionary inside your
    middleware config file.
    """

    def __init__(self, request: Request):
        """Inject Any Dependencies From The Service Container.

        Arguments:
            Request {masonite.request.Request} -- The Masonite request object
        """
        self.request = request

    def before(self):
        """Run This Middleware After The Route Executes."""
        headers = config('middleware.cors') or {}
        self.request.header(headers)



"""Metadate for Masonite package."""
__title__ = 'masonite'
__description__ = 'The core for the Masonite framework'
__url__ = 'https://github.com/MasoniteFramework/masonite'
__version__ = '2.3.0b1.post2'
__author__ = 'Joseph Mancuso'
__author_email__ = 'joe@masoniteproject.com'
__licence__ = 'MIT'



"""Module for the ache disk driver."""

import os

from ...contracts import CacheContract
from ...drivers import BaseCacheDriver
from ...exceptions import DriverLibraryNotFound


class CacheRedisDriver(CacheContract, BaseCacheDriver):
    """Class for the cache redis driver."""

    def __init__(self):
        """Cache redis driver constructor.

        Arguments:
            CacheConfig {config.cache} -- Cache configuration module.
            Application {config.application} -- Application configuration module.
        """
        from config import application, cache

        self.appconfig = application
        self.cache_forever = None
        self.app_name = os.getenv('APP_NAME', 'masonite')

        config = cache.DRIVERS['redis']

        try:
            import redis
            self.redis = redis
        except ImportError:
            raise DriverLibraryNotFound(
                "Could not find the 'redis' library. Run pip install redis to fix this.")

        self.connection = redis.StrictRedis(
            host=config['host'],
            port=config['port'],
            password=config['password'],
            decode_responses=True)

    def store(self, key, value):
        """Stores content in cache file.

        Arguments:
            key {string} -- The key to store the cache file into
            value {string} -- The value you want to store in the cache

        Keyword Arguments:
            extension {string} -- the extension you want to append to the file (default: {".txt"})
            location {string} -- The path you want to store the cache into. (default: {None})

        Returns:
            string -- Returns the key
        """

        self.cache_forever = True

        self.connection.set('{0}_cache_{1}'.format(self.app_name, key), value)

        return key

    def store_for(self, key, value, cache_time, cache_type):
        """Store the cache for a specific amount of time.

        Arguments:
            key {string} -- The key to store the cache file into
            value {string} -- The value you want to store in the cache
            cache_time {int|string} -- The time as a string or an integer (1, 2, 5, 100, etc)
            cache_type {string} -- The type of time to store for (minute, minutes, hours, seconds, etc)

        Keyword Arguments:
            extension {string} -- the extension you want to append to the file (default: {".txt"})
            location {string} -- The path you want to store the cache into. (default: {None})

        Raises:
            ValueError -- Thrown if an invalid cache type was caught (like houes instead of hours).

        Returns:
            string -- Returns the key
        """

        self.cache_forever = False
        cache_for_time = self.calculate_time(cache_type, cache_time)

        self.connection.set('{0}_cache_{1}'.format(self.app_name, key), value, ex=cache_for_time)

        return key

    def get(self, key):
        """Get the data from a key in the cache."""
        return self.connection.get('{0}_cache_{1}'.format(self.app_name, key))

    def delete(self, key):
        """Delete file cache."""
        self.connection.delete('{0}_cache_{1}'.format(self.app_name, key))

    def update(self, key, value):
        """Updates a specific cache by key."""
        time_to_expire = self.connection.ttl('{0}_cache_{1}'.format(self.app_name, key))

        if time_to_expire > 0:
            self.connection.set('{0}_cache_{1}'.format(self.app_name, key), value, ex=time_to_expire)
        else:
            self.connection.set('{0}_cache_{1}'.format(self.app_name, key), value)

        return key

    def exists(self, key):
        """Check if the cache exists."""
        return self.connection.exists('{0}_cache_{1}'.format(self.app_name, key))

    def is_valid(self, key):
        """Check if a valid cache."""
        return self.exists(key)



"""A RouteProvider Service Provider."""

from ..helpers.routes import create_matchurl
from ..provider import ServiceProvider
from ..request import Request
from ..response import Response
from ..routes import Route


class RouteProvider(ServiceProvider):

    def register(self):
        pass

    def boot(self, router: Route, request: Request, response: Response):
        # All routes joined
        from config import application
        for route in self.app.make('WebRoutes'):

            """Make a better match for trailing slashes
            Sometimes a user will end with a trailing slash. Because the user might
            create routes like `/url/route` and `/url/route/` and how the regex
            is compiled down, we may need to adjust for urls that end or dont
            end with a trailing slash.
            """

            matchurl = create_matchurl(router.url, route)

            """Houston, we've got a match
                Check to see if a route matches the corresponding router url. If a match
                is found, execute that route and break out of the loop. We only need
                one match. Routes are executed on a first come, first serve basis
            """
            if matchurl.match(router.url) and request.get_request_method() in route.method_type:
                route.load_request(request)

                """Check if subdomains are active and if the route matches on the subdomain
                    It needs to match to.
                """

                if request.has_subdomain():
                    # Check if the subdomain matches the correct routes domain
                    if not route.has_required_domain():
                        response.view('Route not found. Error 404')
                        continue

                """Get URL Parameters
                    This will create a dictionary of parameters given. This is sort of a short
                    but complex way to retrieve the url parameters.
                    This is the code used to convert /url/@firstname/@lastname
                    to {'firstmane': 'joseph', 'lastname': 'mancuso'}.
                """

                try:
                    parameter_dict = {}
                    for index, value in enumerate(matchurl.match(router.url).groups()):
                        parameter_dict[route.url_list[index]] = value or route.get_default_parameter(route.url_list[index])
                    request.set_params(parameter_dict)
                except AttributeError:
                    pass

                """Excute HTTP before middleware
                    Only those middleware that have a "before" method are ran.
                """

                for http_middleware in self.app.make('HttpMiddleware'):
                    located_middleware = self.app.resolve(
                        http_middleware
                    )
                    if hasattr(located_middleware, 'before'):
                        located_middleware.before()

                """Execute Route Before Middleware
                    This is middleware that contains a before method.
                """

                route.run_middleware('before')

                # Show a helper in the terminal of which route has been visited
                if application.DEBUG:
                    print(request.get_request_method() + ' Route: ' + router.url)

                # If no routes have been found and no middleware has changed the status code
                if not request.get_status():

                    """Get the response from the route and set it on the 'Response' key.
                        This data is typically the output of the controller method depending
                        on the type of route.
                    """

                    response.view(route.get_response(), status=200)

                """Execute Route After Route Middleware
                    This is middleware that contains an after method.
                """

                route.run_middleware('after')

                """Excute HTTP after middleware
                    Only those middleware that have an "after" method are ran.
                    Check here if the middleware even has the required method.
                """

                for http_middleware in self.app.make('HttpMiddleware'):
                    located_middleware = self.app.resolve(http_middleware)

                    if hasattr(located_middleware, 'after'):
                        located_middleware.after()

                """Return breaks the loop because the incoming route is found and executed.
                    There is no need to continue searching the route list. First come
                    first serve around these parts of the woods.
                """
                return

        """No Response was found in the for loop so let's set an arbitrary response now.
        """
        response.view('Route not found. Error 404', status=404)
        # If the route exists but not the method is incorrect
        if request.is_status(404) and request.route_exists(request.path):
            response.view('Method not allowed', status=405)



"""Helper Functions for working with Status Codes."""

statuses = {
    100: '100 Continue',
    101: '101 Switching Protocol',
    102: '102 Processing',
    103: 'Early Hints',
    200: '200 OK',
    201: '201 Created',
    202: '202 Accepted',
    203: '203 Non-Authoritative Information',
    204: '204 No Content',
    205: '205 Reset Content',
    206: '206 Partial Content',
    207: '207 Multi-Status',
    208: '208 Multi-Status',
    226: '226 IM Used',
    300: '300 Multiple Choice',
    301: '301 Moved Permanently',
    302: '302 Found',
    303: '303 See Other',
    304: '304 Not Modified',
    307: '307 Temporary Redirect',
    308: '308 Permanent Redirect',
    400: '400 Bad Request',
    401: '401 Unauthorized',
    402: '402 Payment Required',
    403: '403 Forbidden',
    404: '404 Not Found',
    405: '405 Method Not Allowed',
    406: '406 Not Acceptable',
    407: '407 Proxy Authentication Required',
    408: '408 Request Timeout',
    409: '409 Conflict',
    410: '410 Gone',
    411: '411 Length Required',
    412: '412 Precondition Failed',
    413: '413 Payload Too Large',
    414: '414 URI Too Long',
    415: '415 Unsupported Media Type',
    416: '416 Requested Range Not Satisfiable',
    417: '417 Expectation Failed',
    418: "418 I'm a teapot",
    421: '421 Misdirected Request',
    422: '422 Unprocessable Entity',
    423: '423 Locked',
    424: '424 Failed Dependency',
    425: '425 Too Early',
    426: '426 Upgrade Required',
    428: '428 Precondition Required',
    429: '429 Too Many Requests',
    431: '431 Request Header Fields Too Large',
    451: '451 Unavailable For Legal Reasons',
    500: '500 Internal Server Error',
    501: '501 Not Implemented',
    502: '502 Bad Gateway',
    503: '503 Service Unavailable',
    504: '504 Gateway Timeout',
    505: '505 HTTP Version Not Supported',
    506: '506 Variant Also Negotiates',
    507: '507 Insufficient Storage',
    508: '508 Loop Detected',
    510: '510 Not Extended',
    511: '511 Network Authentication Required',
}


def response_statuses():
    return statuses



from .AppProvider import AppProvider
from .BroadcastProvider import BroadcastProvider
from .CacheProvider import CacheProvider
from .CsrfProvider import CsrfProvider
from .CorsProvider import CorsProvider
from .HelpersProvider import HelpersProvider
from .MailProvider import MailProvider
from .QueueProvider import QueueProvider
from .RouteProvider import RouteProvider
from .SassProvider import SassProvider
from .SessionProvider import SessionProvider
from .UploadProvider import UploadProvider
from .ViewProvider import ViewProvider
from .WhitenoiseProvider import WhitenoiseProvider
from .StatusCodeProvider import StatusCodeProvider



"""Creates New Command Command."""
from ..commands import BaseScaffoldCommand


class CommandCommand(BaseScaffoldCommand):
    """
    Creates a new command.

    command
        {name : Name of the command you would like to create}
    """

    scaffold_name = 'Command'
    postfix = "Command"
    template = '/masonite/snippets/scaffold/command'
    base_directory = 'app/commands/'



"""Queue Manager Module."""

from ..contracts import QueueManagerContract
from .Manager import Manager


class QueueManager(Manager, QueueManagerContract):
    """Manages all queue drivers.

    Arguments:
        Manager {from .managers.Manager} -- The base Manager class.
    """

    config = 'queue'
    driver_prefix = 'Queue'


class Queue:
    """Dummy class that will be used to swap out the manager in the container."""

    pass



"""Authentication Class."""

import uuid

import bcrypt

from ..helpers import password as bcrypt_password, config
from ..app import App


class Auth:
    """This class will be used to authenticate users based on the config/auth.py file."""

    _once = False

    def __init__(self, app: App, auth_model=None):
        """Auth constructor.

        Arguments:
            request {masonite.request.Request} -- The Request object.

        Keyword Arguments:
            auth_model {object} -- The model you want to authenticate with (default: {None})
        """
        self.request = app.make('Request')

        if auth_model:
            self.auth_model = auth_model
        else:
            from config import auth
            self.auth_model = auth.AUTH['model']

        self.driver = config('auth.auth.driver', 'cookie')

    def user(self):
        """Get the currently logged in user.

        Raises:
            exception -- Raised when there has been an error handling the user model.

        Returns:
            object|bool -- Returns the current authenticated user object or False or None if there is none.
        """
        try:
            return self.request.app().make('AuthManager').driver(self.driver).user(self.auth_model)
        except Exception as exception:
            raise exception

        return None

    def login(self, name, password):
        """Login the user based on the parameters provided.

        Arguments:
            name {string} -- The field to authenticate. This could be a username or email address.
            password {string} -- The password to authenticate with.

        Raises:
            exception -- Raised when there has been an error handling the user model.

        Returns:
            object|bool -- Returns the current authenticated user object or False or None if there is none.
        """

        if not isinstance(password, str):
            raise TypeError("Cannot login with password '{}' of type: {}".format(password, type(password)))

        auth_column = self.auth_model.__auth__

        try:
            # Try to login multiple or statements if given an auth list
            if isinstance(auth_column, list):
                model = self.auth_model.where(auth_column[0], name)

                for authentication_column in auth_column[1:]:
                    model.or_where(authentication_column, name)

                model = model.first()
            else:
                model = self.auth_model.where(auth_column, name).first()

            if model and bcrypt.checkpw(bytes(password, 'utf-8'), bytes(self._get_password_column(model), 'utf-8')):
                if not self._once:
                    remember_token = str(uuid.uuid4())
                    model.remember_token = remember_token
                    model.save()
                    self.request.app().make('AuthManager').driver(self.driver).save(remember_token, model=model)
                return model

        except Exception as exception:
            raise exception

        return False

    def logout(self):
        """Logout the current authenticated user.

        Returns:
            self
        """
        self.request.app().make('AuthManager').driver(self.driver).delete()
        return self

    def login_by_id(self, user_id):
        """Login a user by the user ID.

        Arguments:
            user_id {string|int} -- The ID of the user model record.

        Returns:
            object|False -- Returns the current authenticated user object or False or None if there is none.
        """
        model = self.auth_model.find(user_id)

        if model:
            if not self._once:
                remember_token = str(uuid.uuid4())
                model.remember_token = remember_token
                model.save()
                self.request.app().make('AuthManager').driver(self.driver).save(remember_token, model=model)
            return model

        return False

    def once(self):
        """Log in the user without saving a cookie.

        Returns:
            self
        """
        self._once = True
        return self

    def _get_password_column(self, model):
        """Gets the password column to use.

        Arguments:
            model {orator.orm.Model} -- An Orator type model.

        Returns:
            string
        """
        if hasattr(model, '__password__'):
            return getattr(model, model.__password__)

        if hasattr(model, 'password'):
            return getattr(model, 'password')

    def register(self, user):
        """Register the user.

        Arguments:
            user {dict} -- A dictionary of user data information.
        """
        user['password'] = bcrypt_password(user['password'])
        self.auth_model.create(**user)



"""Cache Manager."""

from ..contracts import CacheManagerContract
from .Manager import Manager


class CacheManager(Manager, CacheManagerContract):
    """Manages all cache drivers.

    Arguments:
        Manager {from .managers.Manager} -- The base Manager class.
    """

    config = 'cache'
    driver_prefix = 'Cache'


class Cache:
    """Dummy class that will be used to swap out the manager in the container."""

    pass



"""New Job Command."""
from ..commands import BaseScaffoldCommand


class JobCommand(BaseScaffoldCommand):
    """
    Creates a new Job.

    job
        {name : Name of the job you want to create}
    """

    scaffold_name = 'Job'
    template = '/masonite/snippets/scaffold/job'
    base_directory = 'app/jobs/'



"""A QueueTableCommand Command"""


from cleo import Command

from ..helpers.filesystem import copy_migration


class QueueTableCommand(Command):
    """
    Create migration files for the queue feature

    queue:table
        {--failed : Created the queue failed table}
        {--jobs : Created the queue failed table}
    """

    def handle(self):
        if self.option('failed'):
            copy_migration('masonite/snippets/migrations/create_failed_jobs_table.py')
        if self.option('jobs'):
            copy_migration('masonite/snippets/migrations/create_queue_jobs_table.py')
        self.info('Migration created successfully')



"""A Mail Service Provider."""

from ..drivers import MailMailgunDriver, MailSmtpDriver, \
    MailLogDriver, MailTerminalDriver
from ..managers import MailManager
from ..provider import ServiceProvider
from .. import Mail
from ..helpers import config


class MailProvider(ServiceProvider):

    wsgi = False

    def register(self):
        self.app.bind('MailSmtpDriver', MailSmtpDriver)
        self.app.bind('MailMailgunDriver', MailMailgunDriver)
        self.app.bind('MailLogDriver', MailLogDriver)
        self.app.bind('MailTerminalDriver', MailTerminalDriver)
        self.app.bind('MailManager', MailManager(self.app))

    def boot(self, manager: MailManager):
        self.app.bind('Mail', manager.driver(config('mail.driver')))
        self.app.swap(Mail, manager.driver(config('mail.driver')))



"""A RedirectionProvider Service Provider."""

from ..drivers import SessionCookieDriver, SessionMemoryDriver
from ..managers import SessionManager
from ..provider import ServiceProvider
from ..view import View
from ..request import Request
from .. import Session
from ..helpers import config


class SessionProvider(ServiceProvider):

    def register(self):
        # from config import session
        # self.app.bind('SessionConfig', session)
        self.app.bind('SessionMemoryDriver', SessionMemoryDriver)
        self.app.bind('SessionCookieDriver', SessionCookieDriver)
        self.app.bind('SessionManager', SessionManager(self.app))

    def boot(self, request: Request, view: View, session: SessionManager):
        self.app.bind('Session', session.driver(config('session').DRIVER))
        self.app.swap(Session, session.driver(config('session').DRIVER))
        request.session = self.app.make('Session')

        view.share({
            'session': self.app.make('Session').helper
        })



"""Upload Disk Driver."""

import os
import _io

from ...contracts import UploadContract
from ...drivers import BaseUploadDriver
from ...helpers.filesystem import make_directory


class UploadDiskDriver(BaseUploadDriver, UploadContract):
    """Upload to and from the file system."""

    file_location = None

    def __init__(self):
        """Upload Disk Driver Constructor."""
        pass

    def store(self, fileitem, filename=None, location=None):
        """Store the file onto a server.

        Arguments:
            fileitem {cgi.Storage} -- Storage object.

        Keyword Arguments:
            location {string} -- The location on disk you would like to store the file. (default: {None})
            filename {string} -- A new file name you would like to name the file. (default: {None})

        Returns:
            string -- Returns the file name just saved.
        """

        # use the new filename or get it from the fileitem
        if filename is None:
            filename = self.get_name(fileitem)

        # Check if is a valid extension
        self.validate_extension(self.get_name(fileitem))

        location = self.get_location(location)

        location = os.path.join(location, filename)

        make_directory(location)

        if isinstance(fileitem, _io.TextIOWrapper):
            with open(location, 'wb') as file:
                file.write(bytes(fileitem.read(), 'utf-8'))
        else:
            with open(location, 'wb') as file:
                file.write(fileitem.file.read())

        self.file_location = location

        return filename



"""Publish Service Providers"""
from cleo import Command
from ..helpers import config


class PublishCommand(Command):
    """
    Publishes a Service Provider

    publish
        {name : Name of the Service Provider you want to publish}
        {--t|tag=None : The tag of the provider you want to publish}
    """

    def handle(self):
        from wsgi import container

        for provider in config('providers.providers'):
            if provider.__class__.__name__ == self.argument('name'):
                if self.option('tag') != 'None':
                    provider.publish(tag=self.option('tag'))
                    provider.publish_migrations(tag=self.option('tag'))

                provider.publish()
                provider.publish_migrations()

                return

        raise ValueError('Could not find the {} provider'.format(self.argument('name')))



"""New Key Command."""
from cleo import Command
from cryptography.fernet import Fernet


class KeyCommand(Command):
    """
    Generate a new key.

    key
        {--s|--store : Stores the key in the .env file}
    """

    def handle(self):
        store = self.option('store')
        key = bytes(Fernet.generate_key()).decode('utf-8')

        if store:
            with open('.env', 'r') as file:
                data = file.readlines()

            for line_number, line in enumerate(data):
                if line.startswith('KEY='):
                    data[line_number] = 'KEY={}\n'.format(key)
                    break

            with open('.env', 'w') as file:
                file.writelines(data)

            self.info('Key added to your .env file: {}'.format(key))
        else:
            self.info("Key: {}".format(key))



"""A ModelDocstringCommand Command."""

from cleo import Command


class ModelDocstringCommand(Command):
    """
    Generate a model docstring based on a table definition

    model:docstring
        {table : Name of the table to generate the docstring for}
        {--c|connection=default : The connection to use}
    """
    def handle(self):
        from config.database import DB
        if self.option('connection') == 'default':
            conn = DB.get_schema_manager().list_table_columns(self.argument('table'))
        else:
            conn = DB.connection(self.option('connection')).get_schema_manager().list_table_columns(self.argument('table'))

        docstring = '"""Model Definition (generated with love by Masonite) \n\n'
        for name, column in conn.items():
            length = '({})'.format(column._length) if column._length else ''
            docstring += '{}: {}{} default: {}\n'.format(
                name, column.get_type(), length, column.get_default())

        print(docstring + '"""')



from abc import ABC as Contract
from abc import abstractmethod


class StorageContract(Contract):

    @abstractmethod
    def put(self, location, contents):
        """Puts a file into the correct directory

        Arguments:
            location {string} -- The location of the file
            contents {string|object|file-like object} -- The file object to add.
        """
        pass

    @abstractmethod
    def get(self, location):
        """Get the file contents

        Arguments:
            location {string} -- The location of the file
        """
        pass

    @abstractmethod
    def append(self, location, contents):
        """Get the file contents

        Arguments:
            location {string} -- The location of the file.
            contents {string|object|file-like object} -- The file object to add.
        """
        pass

    @abstractmethod
    def delete(self, location):
        """Deletes the file.

        Arguments:
            location {string} -- The location of the file.
        """
        pass

    @abstractmethod
    def exists(self, location):
        """Checks if a file exists.

        Arguments:
            location {string} -- The location of the file.
        """
        pass

    @abstractmethod
    def driver(self):
        pass

    @abstractmethod
    def url(self, location):
        """Gets the full URL of the file to be served.

        Arguments:
            location {string} -- The location of the file.
        """
        pass

    @abstractmethod
    def size(self, location):
        """Gets the size of the file.

        Arguments:
            location {string} -- The location of the file.
        """
        pass

    @abstractmethod
    def extension(self, location):
        """Gets the extension of the file.

        Arguments:
            location {string} -- The location of the file.
        """
        pass

    @abstractmethod
    def upload(self, *args, **kwargs):
        """Passes all arguments to the upload version of this storage driver.

        Arguments:
            location {string} -- The location of the file.
        """
        pass

    @abstractmethod
    def all(self, location):
        """Gets all files in a specific directory

        Arguments:
            location {string} -- The location of the directory.
        """
        pass

    @abstractmethod
    def make_directory(self, directory):
        """Make an empty directory

        Arguments:
            directory {string} -- The location of the directory.
        """
        pass

    @abstractmethod
    def delete_directory(self, directory, force=False):
        """Delete a directory.

        Arguments:
            directory {string} -- The location of the directory

        Keyword Arguments:
            force {bool} -- Whether or not a directory with contents should be deleted. (default: {False})
        """
        pass

    @abstractmethod
    def move(self, old, new):
        """Move a file from 1 location to another.

        Arguments:
            old {string} -- The file of the file object to be moved.
            new {string} -- The location where the file object should be moved to.
        """
        pass

    @abstractmethod
    def name(self, location):
        """Gets the name of the file with the extension

        Arguments:
            location {string} -- The location of the file
        """
        pass



"""Static Helper Module."""


def static(alias, file_name):
    """Get the static file location of an asset.

    Arguments:
        alias {string} -- The driver and location to search for. This could be s3.uploads
        file_name {string} -- The filename of the file to return.

    Returns:
        string -- Returns the file location.
    """
    from config.storage import DRIVERS
    if '.' in alias:
        alias = alias.split('.')
        location = DRIVERS[alias[0]]['location'][alias[1]]
        if location.endswith('/'):
            location = location[:-1]

        return '{}/{}'.format(location, file_name)

    location = DRIVERS[alias]['location']
    if isinstance(location, dict):
        location = list(location.values())[0]
        if location.endswith('/'):
            location = location[:-1]

    return '{}/{}'.format(location, file_name)



"""An Upload Service Provider."""

from ..drivers import UploadDiskDriver, UploadS3Driver
from ..helpers.static import static
from ..managers import UploadManager
from ..provider import ServiceProvider
from ..view import View
from .. import Upload
from ..helpers import config


class UploadProvider(ServiceProvider):

    wsgi = False

    def register(self):
        # from config import storage
        # self.app.bind('StorageConfig', storage)
        self.app.bind('UploadDiskDriver', UploadDiskDriver)
        self.app.bind('UploadS3Driver', UploadS3Driver)
        self.app.bind('UploadManager', UploadManager(self.app))

    def boot(self, manager: UploadManager, view: View):
        self.app.bind('Upload', manager.driver(config('storage').DRIVER))
        self.app.swap(Upload, manager.driver(config('storage').DRIVER))
        view.share(
            {
                'static': static,
            }
        )



"""Migrate Rollback Command."""
import os
import sys

from cleo import Command
from ..helpers.migrations import Migrations


class MigrateRollbackCommand(Command):
    """
    Migrate Rollback.

    migrate:rollback
        {--c|connection=default : The connection you want to run migrations on}
    """

    def handle(self):
        sys.path.append(os.getcwd())
        migrations = Migrations(self.option('connection')).rollback()
        self.line("")
        for notes in migrations._notes:
            self.line(notes)
        self.line("")



"""Queue Manager Module."""

from ..contracts import StorageManagerContract
from .Manager import Manager


class StorageManager(Manager, StorageManagerContract):
    """Manages all queue drivers.

    Arguments:
        Manager {from .managers.Manager} -- The base Manager class.
    """

    config = 'storage'
    driver_prefix = 'Storage'


class Storage:
    """Dummy class that will be used to swap out the manager in the container."""

    pass



"""The RegisterController Module."""

from config import application
from masonite.auth import Auth
from masonite.request import Request
from masonite.view import View
from masonite.auth import MustVerifyEmail
from masonite.managers import MailManager


class RegisterController:
    """The RegisterController class."""

    def __init__(self):
        """The RegisterController Constructor."""
        pass

    def show(self, request: Request, view: View, auth: Auth):
        """Show the registration page.

        Arguments:
            Request {masonite.request.request} -- The Masonite request class.

        Returns:
            masonite.view.View -- The Masonite View class.
        """
        return view.render('auth/register', {'app': application, 'Auth': auth})

    def store(self, request: Request, mail_manager: MailManager, auth: Auth):
        """Register the user with the database.

        Arguments:
            request {masonite.request.Request} -- The Masonite request class.

        Returns:
            masonite.request.Request -- The Masonite request class.
        """
        user = auth.register({
            'name': request.input('name'),
            'password': request.input('password'),
            'email': request.input('email'),
        })

        if isinstance(user, MustVerifyEmail):
            user.verify_email(mail_manager, request)

        # Login the user
        if auth.login(request.input('email'), request.input('password')):
            # Redirect to the homepage
            return request.redirect('/home')

        # Login failed. Redirect to the register page.
        return request.redirect('/register')



"""List Routes Command."""
from cleo import Command
from tabulate import tabulate


class RoutesCommand(Command):
    """
    List out all routes of the application.

    show:routes
    """

    def handle(self):
        from wsgi import container

        web_routes = container.make('WebRoutes')

        routes = [[
            "Method",
            "Path",
            "Name",
            "Domain",
            "Middleware"
        ]]

        for route in web_routes:
            routes.append([
                route.method_type,
                route.route_url,
                route.named_route,
                route.required_domain,
                ','.join(route.list_middleware),
            ])

        print(tabulate(routes, headers="firstrow", tablefmt="rst"))



"""The ConfirmController Module."""
import datetime

from masonite.auth import Auth, MustVerifyEmail
from masonite.auth.Sign import Sign
from masonite.managers import MailManager
from masonite.request import Request
from masonite.view import View
from masonite.helpers import config


class ConfirmController:
    """The ConfirmController class."""

    def __init__(self):
        """The ConfirmController Constructor."""
        pass

    def verify_show(self, request: Request, view: View, auth: Auth):
        """Show the Verify Email page for unverified users.

        Arguments:
            request {masonite.request.request} -- The Masonite request class.
            request {masonite.view.view} -- The Masonite view class.
            request {masonite.auth.auth} -- The Masonite Auth class.

        Returns:
            [type] -- [description]
        """
        return view.render('auth/verify', {'app': config('application'), 'Auth': auth})

    def confirm_email(self, request: Request, view: View, auth: Auth):
        """Confirm User email and show the correct response.

        Arguments:
            request {masonite.request.request} -- The Masonite request class.
            request {masonite.view.view} -- The Masonite view class.
            request {masonite.auth.auth} -- The Masonite Auth class.

        Returns:
            [type] -- [description]
        """
        sign = Sign()
        token = sign.unsign(request.param('id'))

        if token is not None:
            tokenParts = token.split("::")
            if len(tokenParts) > 1:
                user = auth.auth_model.find(tokenParts[0])

                if user.verified_at is None:
                    timestamp = datetime.datetime.fromtimestamp(float(tokenParts[1]))
                    now = datetime.datetime.now()
                    timestamp_plus_10 = timestamp + datetime.timedelta(minutes=10)

                    if now < timestamp_plus_10:
                        user.verified_at = datetime.datetime.now()
                        user.save()

                        return view.render('auth/confirm', {'app': config('application'), 'Auth': auth})

        return view.render('auth/error', {'app': config('application'), 'Auth': auth})

    def send_verify_email(self, manager: MailManager, request: Request):
        user = request.user()

        if isinstance(user, MustVerifyEmail):
            user.verify_email(manager, request)

        return request.redirect('/home')



"""A Module For Manipulating Code Structures."""

import inspect
import pydoc

from orator.support.collection import Collection as collect


class Dot:

    def dot(self, search, dictionary, default=None):
        """The search string in dot notation to look into the dictionary for.

        Arguments:
            search {string} -- This should be a string in dot notation
                                like 'key.key.value'.
            dictionary {dict} -- A normal dictionary which will be searched using
                                the search string in dot notation.

        Keyword Arguments:
            default {string} -- The default value if nothing is found
                                in the dictionary. (default: {None})

        Returns:
            string -- Returns the value found the dictionary or the default
                        value specified above if nothing is found.
        """
        if '.' not in search:
            if search == '':
                return dictionary
            try:
                return dictionary[search]
            except KeyError:
                return default

        searching = search.split('.')
        possible = None
        while searching:
            dic = dictionary
            for value in searching:
                if not dic:
                    if '*' in searching:
                        return []
                    return default

                if isinstance(dic, list):
                    try:
                        return collect(dic).pluck(searching[searching.index('*') + 1]).serialize()
                    except KeyError:
                        return []

                if not isinstance(dic, dict):
                    return default

                dic = dic.get(value)

                if isinstance(dic, str) and dic.isnumeric():
                    continue

                if dic and not isinstance(dic, int) and len(dic) == 1 and not isinstance(dic[list(dic)[0]], dict):
                    possible = dic

            if not isinstance(dic, dict):
                return dic

            del searching[-1]
        return possible

    def locate(self, search_path, default=''):
        """Locate the object from the given search path

        Arguments:
            search_path {string} -- A search path to fetch the object
                                    from like config.application.debug.

        Keyword Arguments:
            default {string} -- A default string if the search path is
                                not found (default: {''})

        Returns:
            any -- Could be a string, object or anything else that is fetched.
        """
        value = self.find(search_path, default)

        if isinstance(value, dict):
            return self.dict_dot('.'.join(search_path.split('.')[3:]), value, default)

        if value is not None:
            return value

        return default

    def dict_dot(self, search, dictionary, default=''):
        """Takes a dot notation representation of a dictionary and fetches it from the dictionary.

        This will take something like s3.locations and look into the s3 dictionary and fetch the locations
        key.

        Arguments:
            search {string} -- The string to search for in the dictionary using dot notation.
            dictionary {dict} -- The dictionary to search through.

        Returns:
            string -- The value of the dictionary element.
        """
        return self.dot(search, dictionary, default)

    def find(self, search_path, default=''):
        """Used for finding both the uppercase and specified version.

        Arguments:
            search_path {string} -- The search path to find the module,
                                    dictionary key, object etc. This is typically
                                    in the form of dot notation 'config.application.debug'

        Keyword Arguments:
            default {string} -- The default value to return if the search path
                                could not be found. (default: {''})

        Returns:
            any -- Could be a string, object or anything else that is fetched.
        """
        value = pydoc.locate(search_path)

        if value:
            return value

        paths = search_path.split('.')

        value = pydoc.locate('.'.join(paths[:-1]) + '.' + paths[-1].upper())

        if value or value is False:
            return value

        search_path = -1

        # Go backwards through the dot notation until a match is found.
        ran = 0
        while ran < len(paths):
            try:
                value = pydoc.locate('.'.join(paths[:search_path]) + '.' + paths[search_path].upper())
            except IndexError:
                return default

            if value:
                break

            value = pydoc.locate('.'.join(paths[:search_path]) + '.' + paths[search_path])

            if value:
                break

            search_path -= 1
            ran += 1

        if not value or inspect.ismodule(value):
            return default

        return value


def config(path, default=''):
    """Used to fetch a value from a configuration file

    Arguments:
        path {string} -- The search path using dot notation of the value to get

    Keyword Arguments:
        default {str} -- The default value if not value and be found (default: {''})

    Returns:
        mixed
    """
    return Dot().locate('config.' + path, default)


def load(path, default=''):
    """Used to fetch a value from a configuration file

    Arguments:
        path {string} -- The search path using dot notation of the value to get

    Keyword Arguments:
        default {str} -- The default value if not value and be found (default: {''})

    Returns:
        mixed
    """
    return Dot().locate(path, default)



"""Up Command."""

import os

from cleo import Command


class UpCommand(Command):
    """
    Brings the server out of maintenance state.

    up
    """

    def handle(self):
        os.remove('bootstrap/down')



"""Time Module."""

import pendulum


def cookie_expire_time(str_time):
    """Take a string like 1 month or 5 minutes and returns a pendulum instance.

    Arguments:
        str_time {string} -- Could be values like 1 second or 3 minutes

    Returns:
        pendlum -- Returns Pendulum instance
    """
    if str_time != 'expired':
        number = int(str_time.split(" ")[0])
        length = str_time.split(" ")[1]

        if length in ('second', 'seconds'):
            return pendulum.now('GMT').add(seconds=number).format('%a, %d %b %Y %H:%M:%S GMT')
        elif length in ('minute', 'minutes'):
            return pendulum.now('GMT').add(minutes=number).format('%a, %d %b %Y %H:%M:%S GMT')
        elif length in ('hour', 'hours'):
            return pendulum.now('GMT').add(hours=number).format('%a, %d %b %Y %H:%M:%S GMT')
        elif length in ('days', 'days'):
            return pendulum.now('GMT').add(days=number).format('%a, %d %b %Y %H:%M:%S GMT')
        elif length in ('week', 'weeks'):
            return pendulum.now('GMT').add(weeks=1).format('%a, %d %b %Y %H:%M:%S GMT')
        elif length in ('month', 'months'):
            return pendulum.now('GMT').add(months=number).format('%a, %d %b %Y %H:%M:%S GMT')
        elif length in ('year', 'years'):
            return pendulum.now('GMT').add(years=number).format('%a, %d %b %Y %H:%M:%S GMT')

        return None
    else:
        return pendulum.now('GMT').subtract(years=20).format('%a, %d %b %Y %H:%M:%S GMT')


def parse_human_time(str_time):
    """Take a string like 1 month or 5 minutes and returns a pendulum instance.

    Arguments:
        str_time {string} -- Could be values like 1 second or 3 minutes

    Returns:
        pendlum -- Returns Pendulum instance
    """
    if str_time != 'expired':
        number = int(str_time.split(" ")[0])
        length = str_time.split(" ")[1]

        if length in ('second', 'seconds'):
            return pendulum.now('GMT').add(seconds=number)
        elif length in ('minute', 'minutes'):
            return pendulum.now('GMT').add(minutes=number)
        elif length in ('hour', 'hours'):
            return pendulum.now('GMT').add(hours=number)
        elif length in ('days', 'days'):
            return pendulum.now('GMT').add(days=number)
        elif length in ('week', 'weeks'):
            return pendulum.now('GMT').add(weeks=1)
        elif length in ('month', 'months'):
            return pendulum.now('GMT').add(months=number)
        elif length in ('year', 'years'):
            return pendulum.now('GMT').add(years=number)

        return None
    else:
        return pendulum.now('GMT').subtract(years=20)



"""A SassProvider Service Provider."""

import builtins
import os

from ..exception_handler import DD
from ..helpers.view_helpers import back, set_request_method, hidden, old
from ..helpers.sign import sign, unsign, decrypt, encrypt
from ..helpers import config, optional
from ..provider import ServiceProvider
from ..view import View
from ..request import Request
from ..managers import MailManager


class HelpersProvider(ServiceProvider):

    wsgi = False

    def register(self):
        pass

    def boot(self, view: View, request: Request):
        """Add helper functions to Masonite."""
        builtins.view = view.render
        builtins.request = request.helper
        builtins.auth = request.user
        builtins.container = self.app.helper
        builtins.env = os.getenv
        builtins.resolve = self.app.resolve
        builtins.route = request.route
        if self.app.has(MailManager):
            builtins.mail_helper = self.app.make(MailManager).helper
        builtins.dd = DD(self.app).dump

        view.share(
            {
                'request': request.helper,
                'auth': request.user,
                'request_method': set_request_method,
                'route': request.route,
                'back': back,
                'sign': sign,
                'unsign': unsign,
                'decrypt': decrypt,
                'encrypt': encrypt,
                'config': config,
                'optional': optional,
                'dd': builtins.dd,
                'hidden': hidden,
                'exists': view.exists,
                'cookie': request.get_cookie,
                'url': lambda name, params={}: request.route(name, params, full=True),
                'old': old
            }
        )



"""Module for miscellaneous helper methods."""

import random
import string
import warnings

from ..exceptions import AmbiguousError


def random_string(length=4):
    """Generate a random string based on the length given.

    Keyword Arguments:
        length {int} -- The amount of the characters to generate (default: {4})

    Returns:
        string
    """
    return "".join(
        random.choice(string.ascii_uppercase + string.digits) for _ in range(length)
    )


def dot(data, compile_to=None):
    notation_list = data.split('.')

    compiling = ""
    compiling += notation_list[0]
    beginning_string = compile_to.split('{1}')[0]
    compiling = beginning_string + compiling
    dot_split = compile_to.replace(beginning_string + '{1}', '').split('{.}')
    if any(len(x) > 1 for x in dot_split):
        raise ValueError("Cannot have multiple values between {1} and {.}")

    for notation in notation_list[1:]:
        compiling += dot_split[0]
        compiling += notation
        compiling += dot_split[1]
    return compiling


def clean_request_input(value, clean=True):
    if not clean:
        return value

    import html

    try:
        if isinstance(value, str):
            return html.escape(value)
        elif isinstance(value, list):
            return [html.escape(x) for x in value]
        elif isinstance(value, int):
            return value
        elif isinstance(value, dict):
            return {key: html.escape(val) for (key, val) in value.items()}
    except (AttributeError, TypeError):
        pass

    return value


class HasColoredCommands:

    def success(self, message):
        print('\033[92m {0} \033[0m'.format(message))

    def warning(self, message):
        print('\033[93m {0} \033[0m'.format(message))

    def danger(self, message):
        print('\033[91m {0} \033[0m'.format(message))

    def info(self, message):
        return self.success(message)


class Compact:

    def __new__(cls, *args):
        import inspect
        frame = inspect.currentframe()

        cls.dictionary = {}
        for arg in args:

            if isinstance(arg, dict):
                cls.dictionary.update(arg)
                continue

            found = []
            for key, value in frame.f_back.f_locals.items():
                if value == arg:
                    for f in found:
                        if value is f and f is not None:
                            raise AmbiguousError(
                                'Cannot contain variables with multiple of the same object in scope. '
                                'Getting {}'.format(value))
                    cls.dictionary.update({key: value})
                    found.append(value)

        if len(args) != len(cls.dictionary):
            raise ValueError('Could not find all variables in this')
        return cls.dictionary


def deprecated(message):
    def deprecated_decorator(func):
        def deprecated_func(*args, **kwargs):
            warnings.warn("{} is a deprecated function. {}".format(func.__name__, message),
                          category=DeprecationWarning,
                          stacklevel=2)
            warnings.simplefilter('default', DeprecationWarning)
            return func(*args, **kwargs)
        return deprecated_func
    return deprecated_decorator



"""Upload S3 Driver."""

import os

from ...contracts import UploadContract
from ...drivers import BaseUploadDriver
from ...exceptions import DriverLibraryNotFound
from ...managers import UploadManager
from ...app import App
from ...helpers import config


class UploadS3Driver(BaseUploadDriver, UploadContract):
    """Amazon S3 Upload driver."""

    def __init__(self, upload: UploadManager, app: App):
        """Upload Disk Driver Constructor.

        Arguments:
            UploadManager {masonite.managers.UploadManager} -- The Upload Manager object.
            StorageConfig {config.storage} -- Storage configuration.
        """
        self.upload = upload
        self.config = config('storage')

    def store(self, fileitem, filename=None, location=None):
        """Store the file into Amazon S3 server.

        Arguments:
            fileitem {cgi.Storage} -- Storage object.

        Keyword Arguments:
            location {string} -- The location on disk you would like to store the file. (default: {None})
            filename {string} -- A new file name you would like to name the file. (default: {None})

        Raises:
            DriverLibraryNotFound -- Raises when the boto3 library is not installed.

        Returns:
            string -- Returns the file name just saved.
        """
        try:
            import boto3
        except ImportError:
            raise DriverLibraryNotFound(
                'Could not find the "boto3" library. Please pip install this library by running "pip install boto3"')

        driver = self.upload.driver('disk')
        driver.accept_file_types = self.accept_file_types
        driver.store(fileitem, filename=filename, location='storage/temp')
        file_location = driver.file_location

        # use the new filename or get it from the fileitem
        if filename is None:
            filename = self.get_name(fileitem)

        # Check if is a valid extension
        self.validate_extension(self.get_name(fileitem))

        session = boto3.Session(
            aws_access_key_id=self.config.DRIVERS['s3']['client'],
            aws_secret_access_key=self.config.DRIVERS['s3']['secret'],
        )

        s3 = session.resource('s3')

        if location:
            location = os.path.join(location, filename)
        else:
            location = os.path.join(filename)

        s3.meta.client.upload_file(
            file_location,
            self.config.DRIVERS['s3']['bucket'],
            location
        )

        return filename



"""Creates New Test Command."""
from ..commands import BaseScaffoldCommand


class TestCommand(BaseScaffoldCommand):
    """
    Creates a new test case.

    test
        {name : Name of the test you would like to create}
    """

    scaffold_name = 'Test'
    postfix = ""
    prefix = "Test"
    template = '/masonite/snippets/scaffold/test'
    base_directory = 'tests/test_'
    file_to_lower = True



"""Request Module.

Request Module handles many different aspects of a single request
Methods which require the request and help ease development should
be put here.

Methods may return another object if necessary to expand capabilities
of this class.
"""

import re
from cgi import MiniFieldStorage
from http import cookies
from urllib.parse import parse_qsl

import tldextract
from cryptography.fernet import InvalidToken
from .auth.Sign import Sign
from .exceptions import InvalidHTTPStatusCode, RouteException
from .helpers import Dot as DictDot
from .helpers import clean_request_input, dot
from .helpers.Extendable import Extendable
from .helpers.routes import compile_route_to_regex
from .helpers.status import response_statuses
from .helpers.time import cookie_expire_time


class Request(Extendable):
    """Handles many different aspects of a single request.

    This is the object passed through to the controllers
    as a request parameter

    Arguments:
        Extendable {masonite.helpers.Extendable.Extendable} -- Makes this class
        have the ability to extend another class at runtime.
    """

    statuses = response_statuses()

    def __init__(self, environ=None):
        """Request class constructor.

        Initializes several properties and sets various methods
        depending on the environtment.

        Keyword Arguments:
            environ {dictionary} -- WSGI environ dictionary. (default: {None})
        """
        self.cookies = []
        self._headers = {}
        self.url_params = {}
        self.redirect_url = False
        self.redirect_route = False
        self.user_model = None
        self.subdomain = None
        self._activate_subdomains = False
        self._status = None
        self.request_variables = {}
        self._test_user = False

        if environ:
            self.load_environ(environ)

        self.encryption_key = False
        self.container = None

    def input(self, name, default=False, clean=True):
        """Get a specific input value.

        Arguments:
            name {string} -- Key of the input data

        Keyword Arguments:
            default {string} -- Default value if input does not exist (default: {False})
            clean {bool} -- Whether or not the return value should be
                            cleaned (default: {True})

        Returns:
            string
        """
        if '.' in name and isinstance(self.request_variables.get(name.split('.')[0]), dict):
            value = DictDot().dot(name, self.request_variables)
            if value:
                return value

        elif '.' in name:
            name = dot(name, "{1}[{.}]")

        return clean_request_input(self.request_variables.get(name, default), clean=clean)

    def is_post(self):
        """Check if the current request is a POST request.

        Returns:
            bool
        """
        if self.environ['REQUEST_METHOD'] == 'POST':
            return True

        return False

    def is_not_get_request(self):
        """Check if the current request is not a get request.

        Returns:
            bool
        """
        if not self.environ['REQUEST_METHOD'] == 'GET':
            return True

        return False

    def __set_request_method(self):
        """Private method for manually setting the request method.

        Returns:
            bool
        """
        if self.has('__method'):
            self.environ['REQUEST_METHOD'] = self.input('__method')
            return True

        return False

    def key(self, key):
        """Set the encryption key.

        Arguments:
            key {string} -- Encryption key

        Returns:
            self
        """
        self.encryption_key = key
        return self

    def all(self, internal_variables=True, clean=True):
        """Get all the input data.

        Keyword Arguments:
            internal_variables {bool} -- Get the internal framework variables
                                            as well (default: {True})
            clean {bool} -- Whether or not the return value should be
                cleaned (default: {True})

        Returns:
            dict
        """
        if not internal_variables:
            without_internals = {}
            for key, value in self.request_variables.items():
                if not key.startswith('__'):
                    without_internals.update({key: value})
            return clean_request_input(without_internals, clean=clean)

        return clean_request_input(self.request_variables, clean=clean)

    def only(self, *names):
        """Return the specified request variables in a dictionary.

        Returns:
            dict
        """
        only_vars = {}

        for name in names:
            only_vars[name] = self.request_variables.get(name)

        return only_vars

    def without(self, *names):
        """Return the request variables in a dictionary without specified values.

        Returns:
            dict
        """
        only_vars = {}

        for name in self.request_variables:
            if name not in names:
                only_vars[name] = self.request_variables.get(name)

        return only_vars

    def load_app(self, app):
        """Load the container into the request class.

        Arguments:
            app {masonite.app.App} -- Application Container

        Returns:
            self
        """
        self.container = app
        return self

    def load_environ(self, environ):
        """Load the wsgi environment and sets various properties.

        Arguments:
            environ {dict} -- WSGI environ

        Returns:
            self
        """
        self.environ = environ
        self.method = environ['REQUEST_METHOD']
        self.path = environ['PATH_INFO']
        if 'QUERY_STRING' in environ:
            self._set_standardized_request_variables(environ['QUERY_STRING'])

        if self.has('__method'):
            self.__set_request_method()

        return self

    def _set_standardized_request_variables(self, variables):
        """The input data is not perfect so we have to standardize it into a dictionary.

        Arguments:
            variables {string|dict}
        """
        if isinstance(variables, str):
            variables = dict(parse_qsl(variables))

        try:
            self.request_variables = {}
            for name in variables.keys():
                value = self._get_standardized_value(variables[name])
                self.request_variables[name.replace('[]', '')] = value
            return
        except TypeError:
            pass

        self.request_variables = {}

    def _get_standardized_value(self, value):
        """Get the standardized value based on the type of the value parameter.

        Arguments:
            value {list|dict|cgi.FileStorage|string}

        Returns:
            string|bool
        """
        if value is None:
            return None

        if isinstance(value, list):

            # If the list contains MiniFieldStorage objects then loop
            # through and get the values.
            if any(isinstance(storage_obj, MiniFieldStorage) for storage_obj in value):
                values = [storage_obj.value for storage_obj in value]

                # TODO: This needs to be removed in 2.2. A breaking change but
                # this code will result in inconsistent values
                # If there is only 1 element in the list then return the only value in the list
                if len(values) == 1:
                    return values[0]
                return values

            return value

        if isinstance(value, (str, int, dict)):
            return value

        if not value.filename:
            return value.value

        if value.filename:
            return value

        return False

    def app(self):
        """Return the application container.

        Returns:
            masonite.app.App -- Application container
        """
        # if self.container is None:
        #     raise AttributeError("The container has not been loaded into the Request class. Use the 'load_app' method to load the container.")
        return self.container

    def has(self, *args):
        """Check if all given keys in request variable exists.

        Returns:
            bool
        """
        return all((arg in self.request_variables) for arg in args)

    def status(self, status):
        """Set the HTTP status code.

        Arguments:
            status {string|integer} -- A string or integer with the standardized status code

        Returns:
            self
        """
        if isinstance(status, str):
            self.app().bind('StatusCode', status)
        elif isinstance(status, int):
            try:
                text_status = self.statuses[status]
            except KeyError:
                raise InvalidHTTPStatusCode
            self.app().bind('StatusCode', text_status)
        return self

    def get_status_code(self):
        """Return the current request status code.

        Returns:
            string -- Returns the status code (404 Not Found, 200 OK, etc)
        """
        return self.app().make('StatusCode')

    def is_status(self, code):
        return self._get_status_code_by_value(self.get_status_code()) == code

    def route_exists(self, url):
        web_routes = self.container.make('WebRoutes')

        for route in web_routes:
            if route.route_url == url:
                return True

        return False

    def _get_status_code_by_value(self, value):
        for key, status in self.statuses.items():
            if status == value:
                return key

        return None

    def get_status(self):
        return self._get_status_code_by_value(self.get_status_code())

    def get_request_method(self):
        """Get the current request method.

        Returns:
            string -- returns GET, POST, PUT, etc
        """
        return self.environ['REQUEST_METHOD']

    def header(self, key, value=None, http_prefix=None):
        """Set or gets a header depending on if "value" is passed in or not.

        Arguments:
            key {string|dict} -- The header you want to set or get. If the key is a dictionary, loop through each key pair
                                    and add them to the headers.

        Keyword Arguments:
            value {string} -- The value you want to set (default: {None})
            http_prefix {bool} -- Whether it should have `HTTP_` prefixed to the value being set. (default: {True})

        Returns:
            string|None|True -- Either return the value if getting a header,
                                None if it doesn't exist or True if setting the headers.
        """
        if isinstance(key, dict):
            for dic_key, dic_value in key.items():
                self._set_header(dic_key, dic_value, http_prefix)
            return

        # Get Headers
        if value is None:
            if key in self.environ:
                return self.environ[key]
            elif key.upper().replace('-', '_') in self.environ:
                return self.environ[key.upper().replace('-', '_')]
            else:
                return ''

        self._set_header(key, value, http_prefix)

    def _set_header(self, key, value, http_prefix):
        # Set Headers
        if http_prefix:
            self.environ['HTTP_{0}'.format(key)] = str(value)
            self._headers.update({'HTTP_{0}'.format(key): str(value)})
        else:
            self.environ[key] = str(value)
            self._headers.update({key: str(value)})

    def has_raw_header(self, key):
        return key in self._headers

    def get_headers(self):
        """Return all current headers to be set.

        Returns:
            list -- List containing a tuple of headers.
        """

        return self._compile_headers_to_tuple() + self.get_cookies()

    def _compile_headers_to_tuple(self):
        """Compiles the current headers to a list of tuples.

        Returns:
            list -- A list of tuples.
        """

        headers = []
        for key, value in self._headers.items():
            headers.append((key, value))

        return headers

    def reset_headers(self):
        """Reset all headers being set.

        Typically ran at the end of the request
        because of this object acts like a singleton.

        Returns:
            None
        """
        self._headers = {}

    def get_and_reset_headers(self):
        """Gets the headers but resets at the same time.

        This is useful at the end of the WSGI request to prevent
        Several requests from

        Returns:
            tuple
        """
        headers = self.get_headers()
        self.reset_headers()
        self.url_params = {}
        self.cookies = []
        return headers

    def set_params(self, params):
        """Load the params into the class.

        These parameters are where the developer can retrieve the
        /url/@variable:string/ from the url.

        Arguments:
            params {dict} -- Dictionary of parameters to store on the class.

        Returns:
            self
        """
        self.url_params = params
        return self

    def param(self, parameter):
        """Retrieve the param from the URL.

        The "parameter" parameter in this method should be the name of the
        @variable passed into the url in web.py.

        Arguments:
            parameter {string} -- Specific argument to return.

        Returns:
            string|False -- Returns False if key does not exist.
        """
        if parameter in self.url_params:
            return self.url_params.get(parameter)
        return False

    def cookie(self, key, value, encrypt=True,
               http_only="HttpOnly;", path='/', expires=''):
        """Set a cookie in the browser.

        Arguments:
            key {string} -- Name of the cookie you want set.
            value {string} -- Value of the cookie you want set.

        Keyword Arguments:
            encrypt {bool} -- Whether or not you want to encrypt the
                                cookie (default: {True})
            http_only {str} -- If the cookie is HttpOnly or not (default: {"HttpOnly;"})
            path {str} -- The path of the cookie to be set to. (default: {'/'})
            expires {string} -- When the cookie expires
                                (5 minutes, 1 minute, 10 hours, etc) (default: {''})

        Returns:
            self
        """
        if encrypt:
            value = Sign(self.encryption_key).sign(value)
        else:
            value = value

        if expires:
            expires = "Expires={0};".format(cookie_expire_time(expires))

        if not http_only:
            http_only = ""

        self.append_cookie('{0}={1};{2} {3}Path={4}'.format(key, value, expires, http_only, path))
        self.cookies.append(
            ('Set-Cookie', '{0}={1};{2} {3}Path={4}'.format(
                key, value, expires, http_only, path)))
        return self

    def get_cookies(self):
        """Retrieve all cookies from the browser.

        Returns:
            dict -- Returns all the cookies.
        """
        return self.cookies

    def get_raw_cookie(self, provided_cookie):
        if 'HTTP_COOKIE' in self.environ:
            grab_cookie = cookies.SimpleCookie(self.environ['HTTP_COOKIE'])
            if provided_cookie in grab_cookie:
                return grab_cookie[provided_cookie]

        return None

    def get_cookie(self, provided_cookie, decrypt=True):
        """Retrieve a specific cookie from the browser.

        Arguments:
            provided_cookie {string} -- Name of the cookie to retrieve

        Keyword Arguments:
            decrypt {bool} -- Whether Masonite should try to decrypt the cookie.
                              This should only be True if the cookie was encrypted
                              in the first place.  (default: {True})

        Returns:
            string|None -- Returns None if the cookie does not exist.
        """
        if 'HTTP_COOKIE' in self.environ:
            grab_cookie = cookies.SimpleCookie(self.environ['HTTP_COOKIE'])

            if provided_cookie in grab_cookie:
                if decrypt:
                    try:
                        return Sign(self.encryption_key).unsign(
                            grab_cookie[provided_cookie].value)
                    except InvalidToken:
                        self.delete_cookie(provided_cookie)
                        return None

                return grab_cookie[provided_cookie].value
        return None

    def append_cookie(self, value):
        """Append cookie to the string or create a new string.

        Whether a new cookie should append on to the string of cookies to be set
        or create a new string. This string is used by the browser to interpret how
        handle setting a cookie.

        Arguments:
            key {string} -- Name of cookie to be stored
            value {string} -- Value of cookie to be stored
        """
        if 'HTTP_COOKIE' in self.environ and self.environ['HTTP_COOKIE']:
            self.environ['HTTP_COOKIE'] += ';{}'.format(value)
        else:
            self.environ['HTTP_COOKIE'] = '{}'.format(value)

    def delete_cookie(self, key):
        """Delete cookie.

        Arguments:
            key {string} -- Name of cookie to be deleted.

        Returns:
            bool -- Whether or not the cookie was successfully deleted.
        """
        for index, cookie in enumerate(self.cookies):
            if cookie[1].startswith(key + '='):
                del self.cookies[index]

        self.cookie(key, '', expires='expired')

        if 'HTTP_COOKIE' in self.environ and self.environ['HTTP_COOKIE']:

            request_cookies = self.environ['HTTP_COOKIE'].split(';')
            for index, cookie in enumerate(request_cookies):
                if cookie.startswith(key):
                    # remove that cookie
                    del request_cookies[index]

            # put string back together
            self.environ['HTTP_COOKIE'] = ';'.join(request_cookies)
            return True
        return False

    def set_user(self, user_model):
        """Load the user into the class.

        Arguments:
            user_model {app.User.User} -- Defaults to loading this class
                                        unless specifically changed.

        Returns:
            self
        """
        if self._test_user:
            self.user_model = self._test_user
        else:
            self.user_model = user_model

        return self

    def user(self):
        """Load the user into the class.

        Returns:
            app.User.User|None -- Returns None if the user is not loaded or logged in.
        """
        return self.user_model

    def redirect(self, route=None, params={}, name=None, controller=None, status=302):
        """Redirect the user based on the route specified.

        Arguments:
            route {string} -- URI of the route (/dashboard/user)

        Keyword Arguments:
            params {dict} -- Dictionary of parameters to set for the URI.
                             Use this when the URI has something like
                             /dashboard/user/@id. (default: {{}})

        Returns:
            self
        """
        if name:
            return self.redirect_to(name, params, status=status)
        elif route:
            self.redirect_url = self.compile_route_to_url(route, params)
        elif controller:
            self.redirect_url = self.url_from_controller(controller, params)

        self.status(status)
        return self

    def with_input(self):
        self.flash_inputs_to_session()
        return self

    def redirect_to(self, route_name, params={}, status=302):
        """Redirect to a named route.

        Arguments:
            route_name {string} -- Name of a named route.

        Keyword Arguments:
            params {dict} -- Dictionary of parameters to set for the URI.
                             Use this when the URI has something like
                             /dashboard/user/@id. (default: {{}})

        Returns:
            self
        """
        self.redirect_url = self._get_named_route(route_name, params)
        self.status(status)

        return self

    def _get_named_route(self, name, params):
        """Search the list of routes and returns the route with the name passed.

        Arguments:
            name {string} -- Route name to search for (dashboard.user).
            params {dict} -- Dictionary of items to pass to the named route.

        Returns:
            string|None -- Returns None if the route was not found or returns the
                           compiled URI.
        """
        web_routes = self.container.make('WebRoutes')

        for route in web_routes:
            if route.named_route == name:
                return self.compile_route_to_url(route.route_url, params)

        raise RouteException("Could not find the route with the name of '{}'".format(name))

    def _get_route_from_controller(self, controller):
        """Get the route using the controller.

        This finds the route with the attached controller and returns that route.
        This does not compile the URI but actually returns the Route object.

        Arguments:
            controller {string|object} -- Can pass in either a string controller
                                          or the controller itself (the object)

        Returns:
            masonite.routes.Route|None -- Returns None if the route could not be found.
        """
        web_routes = self.container.make('WebRoutes')

        if not isinstance(controller, str):
            module_location = controller.__module__
            controller = controller.__qualname__.split('.')
        else:
            module_location = 'app.http.controllers'
            controller = controller.split('@')

        for route in web_routes:
            if route.controller.__name__ == controller[0] and route.controller_method == controller[1] and route.module_location == module_location:
                return route

    def url_from_controller(self, controller, params={}):
        """Return the compiled URI using a controller.

        Arguments:
            controller {string|object} -- Can be a string controller or
                                            a controller object.

        Keyword Arguments:
            params {dict} -- Dictionary of parameters to pass to the route
                             for compilation. (default: {{}})

        Returns:
            masonite.routes.Route|None -- Returns None if the route could not be found.
        """
        return self.compile_route_to_url(self._get_route_from_controller(controller).route_url, params)

    def route(self, name, params={}, full=False):
        """Get a route URI by its name.

        Arguments:
            name {string} -- Name of the route.

        Keyword Arguments:
            params {dict} -- Dictionary of parameters to pass to the route
                             for compilation. (default: {{}})
            full {bool} -- Specifies whether the full application url should
                           be returned or not. (default: {False})

        Returns:
            masonite.routes.Route|None -- Returns None if the route cannot be found.
        """
        from config import application
        if full:
            route = application.URL + self._get_named_route(name, params)
        else:
            try:
                route = self._get_named_route(name, params)
            except KeyError:
                params = {}
                params.update(self.url_params)
                route = self._get_named_route(name, params)

        if not route:
            raise RouteException("Route with the name of '{}' was not found.".format(name))

        return route

    def __getattr__(self, key):
        inp = self.input(key)
        if inp:
            return inp

        inp = self.param(key)
        if inp:
            return inp

        raise AttributeError("class 'Request' has no attribute {}".format(key))

    def with_errors(self, errors):
        self.session.flash('errors', errors)
        return self

    def reset_redirections(self):
        """Reset the redirections because of this class acting like a singleton pattern."""
        self.redirect_url = False
        self.redirect_route = False

    def back(self, default=None):
        """Return a URI for redirection depending on several use cases.

        Keyword Arguments:
            default {string} -- Default value if nothing can be found. (default: {None})

        Returns:
            self
        """
        self.with_input()

        redirect_url = self.input('__back')
        if not redirect_url and default:
            return self.redirect(default)
        elif not redirect_url and not default:
            return self.redirect(self.path)  # Some global default?

        return self.redirect(redirect_url)

    def flash_inputs_to_session(self):
        if not hasattr(self, 'session'):
            return

        for key, value in self.all().items():
            self.session.flash(key, value)

    def is_named_route(self, name, params={}):
        """Check if the current URI is a specific named route.

        Arguments:
            name {string} -- The name of a route.

        Keyword Arguments:
            params {dict} -- Dictionary of parameters to pass to the route. (default: {{}})

        Returns:
            bool
        """
        if self._get_named_route(name, params) == self.path:
            return True

        return False

    def contains(self, route, show=None):
        """If the specified URI is in the current URI path.

        Arguments:
            route {string} -- Part of a URI (/dashboard)

        Returns:
            bool
        """
        if show is not None:
            if re.match(compile_route_to_regex(route), self.path):
                return show

            return ''

        return re.match(compile_route_to_regex(route), self.path)

    def compile_route_to_url(self, route, params={}):
        """Compile the route url into a usable url.

        Converts /url/@id into /url/1. Used for redirection

        Arguments:
            route {string} -- An uncompiled route
                                like (/dashboard/@user:string/@id:int)

        Keyword Arguments:
            params {dict} -- Dictionary of parameters to pass to the route (default: {{}})

        Returns:
            string -- Returns a compiled string (/dashboard/joseph/1)
        """
        if "http" in route:
            return route

        # Split the url into a list
        split_url = route.split('/')

        # Start beginning of the new compiled url
        compiled_url = '/'

        # Iterate over the list
        for url in split_url:
            if url:
                # if the url contains a parameter variable like @id:int
                if '@' in url:
                    url = url.replace('@', '').split(':')[0]
                    if isinstance(params, dict):
                        compiled_url += str(params[url]) + '/'
                    elif isinstance(params, list):
                        compiled_url += str(params.pop(0)) + '/'
                elif '?' in url:
                    url = url.replace('?', '').split(':')[0]
                    if isinstance(params, dict):
                        compiled_url += str(params.get(url, '/')) + '/'
                    elif isinstance(params, list):
                        compiled_url += str(params.pop(0)) + '/'
                else:
                    compiled_url += url + '/'

        compiled_url = compiled_url.replace('//', '')
        # The loop isn't perfect and may have an unwanted trailing slash
        if compiled_url.endswith('/') and not route.endswith('/'):
            compiled_url = compiled_url[:-1]

        # The loop isn't perfect and may have 2 slashes next to eachother
        if '//' in compiled_url:
            compiled_url = compiled_url.replace('//', '/')

        return compiled_url

    def activate_subdomains(self):
        """Activate subdomains abilities."""
        self._activate_subdomains = True

    def has_subdomain(self):
        """Check if the current URI has a subdomain.

        Returns:
            bool
        """
        if self._activate_subdomains:
            url = tldextract.extract(self.environ['HTTP_HOST'])

            if url.subdomain:
                self.subdomain = url.subdomain
                self.url_params.update({'subdomain': self.subdomain})
                return True

        return False

    def send(self, params):
        """DEPRECATED :: sets a dictionary to be compiled for a route.

        Arguments:
            params {dict} -- Dictionary of parameters you want to pass to the route.

        Returns:
            self
        """
        self.set_params(params)
        return self

    def helper(self):
        """Dummy method to work with returning the class. Used for helper methods in the View class.

        Returns:
            self
        """
        return self

    def pop(self, *input_variables):
        """Delete keys from the request input."""
        for key in input_variables:
            if key in self.request_variables:
                del self.request_variables[key]

    def validate(self, *rules):
        validator = self.app().make('Validator')
        return validator.validate(self.request_variables, *rules)



"""Run Seed Command."""
import subprocess

from cleo import Command


class SeedRunCommand(Command):
    """
    Run seed for database.

    seed:run
        {table=None : Name of the table to seed}
    """

    def handle(self):

        table = self.argument('table').lower()
        if not table == 'none':
            seeder = '--seeder {}_table_seeder'.format(table.lower())
        else:
            seeder = ''

        subprocess.call([
            "orator db:seed -p databases/seeds -c config/database.py -f {}".format(
                seeder),
        ], shell=True)



"""AuthJWTDriver Module."""

import pendulum
from ...auth import Auth
from ...contracts import AuthContract
from ...drivers import BaseDriver
from ...exceptions import DriverLibraryNotFound
from ...helpers import config, cookie_expire_time
from ...request import Request


class AuthJwtDriver(BaseDriver, AuthContract):

    def __init__(self, request: Request):
        """AuthCookieDriver initializer.

        Arguments:
            request {masonite.request.Request} -- The Masonite request class.
        """
        self.request = request
        try:
            import jwt
            self.jwt = jwt
        except ImportError:
            raise DriverLibraryNotFound("Please install pyjwt by running 'pip install pyjwt'")

    def user(self, auth_model):
        """Gets the user based on this driver implementation

        Arguments:
            auth_model {orator.orm.Model} -- An Orator ORM type object.

        Returns:
            Model|bool
        """
        from config.application import KEY
        if self.request.get_cookie('token'):

            try:
                token = self.jwt.decode(self.request.get_cookie('token'), KEY, algorithms=['HS256'])
            except self.jwt.exceptions.DecodeError:
                self.delete()
                return False

            expired = token['expired']
            token.pop('expired')
            if not pendulum.parse(expired).is_past():
                auth_model = auth_model()
                auth_model.fill(**token)
                return auth_model
            else:
                if config('auth.drivers.jwt.reauthentication', True):
                    auth_model = Auth(self.request).login_by_id(token[auth_model.__primary_key__])
                else:
                    auth_model.fill(**token)

                token.update({
                    'expired': cookie_expire_time(config('auth.drivers.jwt.lifetime', '5 minutes'))
                })
                self.request.cookie('token', token)
                return auth_model
        return False

    def save(self, _, **kwargs):
        """Saves the state of authentication.

        In this case the state is saving to a cookie.

        Arguments:
            remember_token {string} -- A token containing the state.

        Returns:
            bool
        """
        from config.application import KEY
        model = kwargs.get('model', False)
        serialized_dictionary = model.serialize()
        serialized_dictionary.update({
            'expired': cookie_expire_time('5 minutes')
        })
        token = self.jwt.encode(serialized_dictionary, KEY, algorithm='HS256')
        token = bytes(token).decode('utf-8')
        self.request.cookie('token', token)

    def delete(self):
        """Deletes the state depending on the implementation of this driver.

        Returns:
            bool
        """
        self.request.delete_cookie('token')



"""New Model Command."""
from ..app import App
from ..helpers.filesystem import make_directory
from ..view import View

from cleo import Command


class ModelCommand(Command):
    """
    Creates a model.

    model
        {name : Name of the model}
        {--m|migration : Create a migration for specified model}
        {--s|seed=? : Create a database seed}
    """

    scaffold_name = "Model"
    template = '/masonite/snippets/scaffold/model'
    base_directory = 'app/'

    def handle(self):
        class_name = self.argument('name')
        view = View(App())
        class_directory = '{}{}.py'.format(self.base_directory, class_name)

        if not make_directory(class_directory):
            return self.error('{0} Already Exists!'.format(self.scaffold_name))

        with open(class_directory, 'w+') as f:
            if view.exists(self.template):
                f.write(
                    view.render(self.template, {
                                'class': class_name.split('/')[-1]}).rendered_template
                )
                self.info('{} Created Successfully!'.format(self.scaffold_name))

        if self.option('migration'):
            model_name = class_name.lower() + 's'
            self.call('migration', [
                ('name', 'create_{}_table'.format(model_name)),
                ('-c', model_name)
            ])

        if self.option('seed'):
            seed_file = model_name
            seed_file = self.option('seed')

            self.call('seed', [
                ('table', seed_file)
            ])



from ..auth import Sign


def sign(value):
    """Shortcut for Sign class.

    Arguments:
        value {string} -- The value that is going to be encrypted

    Returns:
        string -- The string value after encryption.
    """
    return Sign().sign(value)


def encrypt(value):
    """Shortcut for Sign class sign method.

    Arguments:
        value {string} -- The value that is going to be encrypted

    Returns:
        string -- The string value after encryption.
    """
    return sign(value)


def unsign(value):
    """Shortcut for Sign class unsign method.

    Arguments:
        value {string} -- The value that is going to be decrypted

    Returns:
        string -- The string value after decryption.
    """
    return Sign().unsign(value)


def decrypt(value):
    """Shortcut for Sign class unsign method.

    Arguments:
        value {string} -- The value that is going to be decrypted

    Returns:
        string -- The string value after decryption.
    """
    return unsign(value)



from ..provider import ServiceProvider
from ..request import Request
from ..response import Response
from ..helpers import config


class CorsProvider(ServiceProvider):
    """Provides Services To The Service Container."""

    wsgi = True

    def register(self):
        """Register objects into the Service Container."""
        pass

    def boot(self, request: Request, response: Response):
        """Boots services required by the container."""
        headers = config('middleware.cors') or {}
        request.header(headers)

        if request.get_request_method().lower() == 'options':
            response.view('preflight')



from cleo import Command

from ..app import App
from ..helpers.filesystem import make_directory
from ..view import View


class BaseScaffoldCommand(Command):
    """
    Creates a model.

    model
        {name : Name of the model}
    """

    scaffold_name = 'Example'
    suffix = ''
    postfix = ''
    prefix = ''
    file_extension = '.py'
    base_directory = 'app/example/'
    file_to_lower = False

    template = '/masonite/snippets/scaffold/model'

    def handle(self):
        class_name = self.argument('name') + self.postfix
        view = View(App())
        class_directory = '{}{}{}{}'.format(
            self.base_directory, class_name, self.suffix, self.file_extension)

        if self.file_to_lower:
            class_directory = class_directory.lower()

        if not make_directory(class_directory):
            return self.error('{0} Already Exists!'.format(self.scaffold_name))

        with open(class_directory, 'w+') as f:
            if view.exists(self.template):
                f.write(
                    view.render(self.template, {
                                'class': self.prefix + class_name.split('/')[-1]}).rendered_template
                )
                self.info('{} Created Successfully!'.format(self.scaffold_name))



"""Down Command."""

from cleo import Command


class DownCommand(Command):
    """
    Puts the server in a maintenance state.

    down
    """

    def handle(self):
        with open('bootstrap/down', 'w+'):
            pass



"""A Cache Service Provider."""

from .. import Cache
from ..drivers import CacheDiskDriver, CacheRedisDriver
from ..managers import CacheManager
from ..provider import ServiceProvider
from ..helpers import config


class CacheProvider(ServiceProvider):

    wsgi = False

    def register(self):
        # from config import cache
        # self.app.bind('CacheConfig', cache)
        self.app.bind('CacheDiskDriver', CacheDiskDriver)
        self.app.bind('CacheRedisDriver', CacheRedisDriver)
        self.app.bind('CacheManager', CacheManager(self.app))

    def boot(self, cache: CacheManager):
        self.app.bind('Cache', cache.driver(config('cache').DRIVER))
        self.app.swap(Cache, cache.driver(config('cache').DRIVER))



"""A View Service Provider."""

from jinja2 import FileSystemLoader

from ..provider import ServiceProvider
from ..view import View


class ViewProvider(ServiceProvider):

    wsgi = False

    def register(self):
        view = View(self.app)
        self.app.bind('ViewClass', view)
        self.app.bind('View', view.render)

    def boot(self, view: View):
        view.add_environment('src/masonite/snippets', loader=FileSystemLoader)
        self.publishes_migrations([
            'storage/append_from.txt'
        ])



import io

from ..app import App
from .TestRequest import TestRequest
from .TestRoute import TestRoute


def generate_wsgi():
    return {
        'wsgi.version': (1, 0),
        'wsgi.multithread': False,
        'wsgi.multiprocess': True,
        'wsgi.run_once': False,
        'wsgi.input': io.BytesIO(),
        'SERVER_SOFTWARE': 'gunicorn/19.7.1',
        'REQUEST_METHOD': 'GET',
        'QUERY_STRING': 'application=Masonite',
        'RAW_URI': '/',
        'SERVER_PROTOCOL': 'HTTP/1.1',
        'HTTP_HOST': '127.0.0.1:8000',
        'HTTP_ACCEPT': 'text/html,application/xhtml+xml,application/xml;q=0.9,*/*;q=0.8',
        'HTTP_UPGRADE_INSECURE_REQUESTS': '1',
        'HTTP_COOKIE': 'setcookie=value',
        'HTTP_USER_AGENT': 'Mozilla/5.0 (Macintosh; Intel Mac OS X 10_13_2) AppleWebKit/604.4.7 (KHTML, like Gecko) Version/11.0.2 Safari/604.4.7',
        'HTTP_ACCEPT_LANGUAGE': 'en-us',
        'HTTP_ACCEPT_ENCODING': 'gzip, deflate',
        'HTTP_CONNECTION': 'keep-alive',
        'wsgi.url_scheme': 'http',
        'REMOTE_ADDR': '127.0.0.1',
        'REMOTE_PORT': '62241',
        'SERVER_NAME': '127.0.0.1',
        'SERVER_PORT': '8000',
        'PATH_INFO': '/',
        'SCRIPT_NAME': ''
    }


class TestSuite:

    def create_container(self, wsgi=generate_wsgi(), container=None, routes=[]):
        from config import providers
        if not container:
            container = App(remember=False)

        container.bind('WSGI', wsgi)
        container.bind('Container', container)
        container.bind('ProviderList', providers.PROVIDERS)
        container.bind('Providers', [])
        container.bind('WSGIProviders', [])

        """
        |--------------------------------------------------------------------------
        | Bind all service providers
        |--------------------------------------------------------------------------
        |
        | Let's register everything into the Service Container. Once everything is
        | in the container we can run through all the boot methods. For reasons
        | some providers don't need to execute with every request and should
        | only run once when the server is started. Providers will be ran
        | once if the wsgi attribute on a provider is False.
        |
        """

        for provider in container.make('ProviderList'):
            try:
                located_provider = provider()
            except TypeError:
                located_provider = provider

            located_provider.load_app(container).register()
            if located_provider.wsgi:
                container.make('WSGIProviders').append(located_provider)
            else:
                container.make('Providers').append(located_provider)

        for provider in container.make('Providers'):
            container.resolve(provider.boot)

        """
        |--------------------------------------------------------------------------
        | Startup the Service Container
        |--------------------------------------------------------------------------
        |
        | Instantiate the Service Container so we can bind classes into it and
        | bind the environ variable that is created by the WSGI server into
        | the container.
        |
        """

        container.bind('Environ', wsgi)

        """
        |--------------------------------------------------------------------------
        | Execute All Service Providers
        |--------------------------------------------------------------------------
        |
        | Run all service provider boot methods if the wsgi attribute is true.
        |
        """

        if routes:
            container.bind('WebRoutes', routes)

        for provider in container.make('WSGIProviders'):
            container.resolve(provider.boot)

        self.container = container
        return self

    def get_container(self):
        return self.container

    def route(self, route):
        return TestRoute(route)

    def get(self, route_url):
        return TestRequest().get(route_url)



from abc import ABC, abstractmethod


class MailContract(ABC):

    @abstractmethod
    def to(self):
        pass

    @abstractmethod
    def template(self):
        pass

    @abstractmethod
    def send_from(self):
        pass

    @abstractmethod
    def subject(self):
        pass



"""Driver for AMQP support"""

import inspect
import pickle
import time

import pendulum
from ...contracts import QueueContract
from ...drivers import BaseQueueDriver
from ...exceptions import DriverLibraryNotFound
from ...helpers import HasColoredCommands
from ...queues import Queueable


class QueueAmqpDriver(BaseQueueDriver, QueueContract, HasColoredCommands):

    def __init__(self):
        """Queue AMQP Driver."""
        from config import queue
        self.queue = queue
        if 'amqp' in self.queue.DRIVERS:
            listening_channel = self.queue.DRIVERS['amqp']['channel']
        else:
            listening_channel = 'default'

        # Start the connection
        self.publishing_channel = listening_channel
        self.connect()

    def _publish(self, body):

        self.channel.basic_publish(exchange='',
                                   routing_key=self.publishing_channel,
                                   body=pickle.dumps(
                                       body
                                   ),
                                   properties=self.pika.BasicProperties(
                                       delivery_mode=2,  # make message persistent
                                   ))
        self.channel.close()
        self.connection.close()

    def push(self, *objects, args=(), callback='handle', ran=1, channel=None, **options):
        """Push objects onto the amqp stack.

        Arguments:
            objects {*args of objects} - This can be several objects as parameters into this method.
        """
        if channel:
            self.publishing_channel = channel

        for obj in objects:
            # Publish to the channel for each object
            payload = {'obj': obj, 'args': args, 'callback': callback, 'created': pendulum.now(), 'ran': ran}
            try:
                additional_exceptions = (self.pika.exceptions.ConnectionWrongStateError, self.pika.exceptions.ChannelWrongStateError)
            except AttributeError:
                additional_exceptions = ()

            try:
                self._publish(payload)
            except ((self.pika.exceptions.ConnectionClosed, self.pika.exceptions.ChannelClosed), additional_exceptions):
                self.connect()
                self._publish(payload)

    def connect(self):
        try:
            import pika
            self.pika = pika
        except ImportError:
            raise DriverLibraryNotFound(
                "Could not find the 'pika' library. Run pip install pika to fix this.")

        self.connection = pika.BlockingConnection(pika.URLParameters('amqp://{}:{}@{}{}/{}'.format(
            self.queue.DRIVERS['amqp']['username'],
            self.queue.DRIVERS['amqp']['password'],
            self.queue.DRIVERS['amqp']['host'],
            ':' + str(self.queue.DRIVERS['amqp']['port']) if 'port' in self.queue.DRIVERS['amqp'] and self.queue.DRIVERS['amqp']['port'] else '',
            self.queue.DRIVERS['amqp']['vhost'] if 'vhost' in self.queue.DRIVERS['amqp'] and self.queue.DRIVERS['amqp']['vhost'] else '%2F'
        )))

        self.channel = self.connection.channel()

        self.channel.queue_declare(self.publishing_channel, durable=True)

        return self

    def consume(self, channel, fair=False):
        self.success('[*] Waiting to process jobs on the "{}" channel. To exit press CTRL+C'.format(
            channel))

        if fair:
            self.channel.basic_qos(prefetch_count=1)

        self.basic_consume(self.work, channel)

        try:
            self.channel.start_consuming()
        finally:
            self.channel.stop_consuming()
            self.channel.close()
            self.connection.close()

    def basic_consume(self, callback, queue_name):
        try:
            self.channel.basic_consume(callback, queue=queue_name)
        except TypeError:
            self.channel.basic_consume(queue_name, callback)

    def work(self, ch, method, properties, body):
        from wsgi import container
        job = pickle.loads(body)
        obj = job['obj']
        args = job['args']
        callback = job['callback']
        ran = job['ran']

        try:
            try:
                if inspect.isclass(obj):
                    obj = container.resolve(obj)

                getattr(obj, callback)(*args)

            except AttributeError:
                obj(*args)

            try:
                self.success('[\u2713] Job Successfully Processed')
            except UnicodeEncodeError:
                self.success('[Y] Job Successfully Processed')
        except Exception as e:
            self.danger('Job Failed: {}'.format(str(e)))

            if not obj.run_again_on_fail:
                ch.basic_ack(delivery_tag=method.delivery_tag)
                return

            if ran < obj.run_times and isinstance(obj, Queueable):
                time.sleep(1)
                self.push(obj.__class__, args=args, callback=callback, ran=ran + 1)
            else:
                if hasattr(obj, 'failed'):
                    getattr(obj, 'failed')(job, str(e))

                self.add_to_failed_queue_table(job)

        ch.basic_ack(delivery_tag=method.delivery_tag)



"""A QueueWorkCommand Command."""

from cleo import Command

from .. import Queue


class QueueWorkCommand(Command):
    """
    Start the queue worker

    queue:work
        {--c|channel=default : The channel to listen on the queue}
        {--d|driver=default : Specify the driver you would like to connect to}
        {--f|fair : Send jobs to queues that have no jobs instead of randomly selecting a queue}
        {--failed : Run only the failed jobs}
    """

    def handle(self):
        from wsgi import container

        if self.option('driver') == 'default':
            queue = container.make(Queue)
        else:
            queue = container.make(Queue).driver(self.option('driver'))

        if self.option('failed'):
            queue.run_failed_jobs()
            return

        queue.connect().consume(self.option('channel'), fair=self.option('fair'))



from orator.migrations import Migration


class CreateQueueJobsTable(Migration):

    def up(self):
        """Run the migrations."""
        with self.schema.create('queue_jobs') as table:
            table.increments('id')
            table.string('name')
            table.binary('serialized')
            table.integer('attempts')
            table.integer('failed').nullable()
            table.timestamp('ran_at').nullable()
            table.timestamp('created_at').nullable()
            table.timestamp('wait_until').nullable()

    def down(self):
        """Revert the migrations."""
        self.schema.drop('queue_jobs')



"""Session Manager Module."""

from ..contracts import SessionManagerContract
from .Manager import Manager


class SessionManager(Manager, SessionManagerContract):
    """Manages all session drivers.

    Arguments:
        Manager {from .managers.Manager} -- The base Manager class.
    """

    config = 'session'
    driver_prefix = 'Session'


class Session:
    """Dummy class that will be used to swap out the manager in the container."""

    pass



from .static import static
from .password import password
from .misc import random_string, dot, clean_request_input, HasColoredCommands, Compact as compact, deprecated
from .Extendable import Extendable
from .time import cookie_expire_time, parse_human_time
from .optional import Optional as optional
from .structures import config, Dot, load
from .migrations import has_unmigrated_migrations
from orator.support.collection import Collection as collect



"""A StatusProvider Service Provider."""

from ..response import Response
from ..provider import ServiceProvider
from ..helpers import config

class ServerErrorExceptionHook:

    def load(self, app):
        from config import application
        if application.DEBUG:
            return

        request = app.make('Request')

        request.status(500)
        if app.make('ViewClass').exists('errors/500'):
            rendered_view = app.make('View')('errors/500')
        else:
            rendered_view = app.make('View')(
                config('application.templates.statuscode', '/masonite/snippets/statuscode'), {'code': '500 Internal Server Error'})

        request.app().make(Response).view(rendered_view)


class StatusCodeProvider(ServiceProvider):

    def register(self):
        self.app.bind('ServiceErrorExceptionHook', ServerErrorExceptionHook())

    def boot(self):
        request = self.app.make('Request')
        response = self.app.make(Response)
        if request.is_status(200):
            return

        if request.get_status() in (500, 405, 404):
            if request.header('Content-Type') == 'application/json':
                # Returns json response when we want the client to receive a json response
                json_response = {'error': {'status': request.get_status()}}
                response.view(json_response, status=request.get_status())
            else:
                # Returns html response when json is not explicitly specified
                if self.app.make('ViewClass').exists('errors/{}'.format(request.get_status())):
                    rendered_view = self.app.make('View')(
                        'errors/{}'.format(request.get_status()))
                else:
                    rendered_view = self.app.make('View')(config('application.templates.statuscode', '/masonite/snippets/statuscode'), {
                        'code': request.get_status_code()
                    })

                response.view(rendered_view, status=request.get_status())



"""The Masonite Response Object."""

import json

from .exceptions import ResponseError
from .helpers.Extendable import Extendable

from orator.support.collection import Collection
from orator import Model
from .app import App
from orator import Paginator, LengthAwarePaginator


class Response(Extendable):

    def __init__(self, app: App):
        """A Response object to be used to abstract the logic of getting a response ready to be returned.

        Arguments:
            app {masonite.app.App} -- The Masonite container.
        """
        self.app = app
        self.request = self.app.make('Request')

    def json(self, payload, status=200):
        """Gets the response ready for a JSON response.

        Arguments:
            payload {dict|list} -- Either a dictionary or a list.

        Returns:
            string -- Returns a string representation of the data
        """
        self.app.bind('Response', json.dumps(payload))
        self.make_headers(content_type="application/json; charset=utf-8")
        self.request.status(status)

        return self.data()

    def paginated_json(self, paginator, status=200):
        """Determine type of paginated instance and return JSON response.

        Arguments:
            paginator {Paginator|LengthAwarePaginator} --
                Either an Orator Paginator or LengthAwarePaginator object

        Returns:
            string -- Returns a string representation of the data
        """
        # configured param types
        page_size_parameter = 'page_size'
        page_parameter = 'page'

        # try to capture request input for page_size and/or page
        page_size_input = self.request.input(page_size_parameter)
        page_input = self.request.input(page_parameter)
        # use try/except here, as int(bool) will return 0 for False above
        try:
            page_size = (
                int(page_size_input)
                if page_size_input and int(page_size_input) > 0
                else paginator.per_page
            )
        except Exception:
            page_size = paginator.per_page
        try:
            page = (
                int(page_input)
                if page_input and int(page_input) > 0
                else paginator.current_page
            )
        except Exception:
            page = paginator.current_page

        # don't waste time instantiating new paginator if no change
        if (
            page_size != paginator.per_page
            or page != paginator.current_page
        ):
            try:
                # try to get class of model
                next(type(x) for x in paginator.items)
                if isinstance(paginator, Paginator):
                    paginator = model_class.simple_paginate(
                        page_size,
                        page
                    )
                elif isinstance(paginator, LengthAwarePaginator):
                    paginator = model_class.paginate(page_size, page)
            except Exception:
                paginator = paginator

        payload = {
            'total': (
                paginator.total
                if isinstance(paginator, LengthAwarePaginator)
                else None
            ),
            'count': paginator.count(),
            'per_page': page_size,
            'current_page': page,
            'last_page': (
                paginator.last_page
                if isinstance(paginator, LengthAwarePaginator)
                else None
            ),
            'from': (page_size * (page - 1)) + 1,
            'to': page_size * page,
            'data': paginator.serialize()
        }

        # remove fields not relevant to Paginator instance
        if isinstance(paginator, Paginator):
            del payload['total']
            del payload['last_page']

        return self.json(payload, status)

    def make_headers(self, content_type="text/html; charset=utf-8"):
        """Make the appropriate headers based on changes made in controllers or middleware.

        Keyword Arguments:
            content_type {str} -- The content type to set. (default: {"text/html; charset=utf-8"})
        """
        self.request.header('Content-Length', str(len(self.to_bytes())))

        # If the user did not change it directly
        if not self.request.has_raw_header('Content-Type'):
            self.request.header('Content-Type', content_type)

    def data(self):
        """Get the data that will be returned to the WSGI server.

        Returns:
            string -- Returns a string representation of the response
        """
        if self.app.has('Response'):
            return self.app.make('Response')

        return ''

    def converted_data(self):
        """Converts the data appropriately so the WSGI server can handle it.

        Returns:
            string -- Returns a string representation of the data
        """
        if isinstance(self.data(), (dict, list)):
            return json.dumps(self.data())
        else:
            return self.data()

    def view(self, view, status=200):
        """Set a string or view to be returned.

        Arguments:
            view {string|dict|list|masonite.view.View} -- Some data type that is an appropriate response.

        Keyword Arguments:
            status {int} -- The Response status code. (default: {200})

        Raises:
            ResponseError -- If a data type that is not an acceptable response type is returned.

        Returns:
            string|dict|list -- Returns the data to be returned.
        """
        if not self.request.get_status():
            self.request.status(status)

        if isinstance(view, (dict, list)):
            return self.json(view, status=self.request.get_status())
        elif isinstance(view, (Collection, Model)):
            return self.json(view.serialize(), status=self.request.get_status())
        elif isinstance(view, int):
            view = str(view)
        elif isinstance(view, Responsable):
            view = view.get_response()
        elif isinstance(view, self.request.__class__):
            view = self.data()
        if isinstance(view, (Paginator, LengthAwarePaginator)):
            return self.paginated_json(view, status=self.request.get_status())
        elif view is None:
            raise ResponseError('Responses cannot be of type: None. Did you return anything in your responsable method?')

        if not isinstance(view, str):
            raise ResponseError('Invalid response type of {}'.format(type(view)))

        self.app.bind('Response', view)

        self.make_headers()

        return self.data()

    def redirect(self, location=None, status=302):
        """Set the redirection on the server.

        Keyword Arguments:
            location {string} -- The URL to redirect to (default: {None})
            status {int} -- The Response status code. (default: {302})

        Returns:
            string -- Returns the data to be returned.
        """
        self.request.status(status)
        if not location:
            location = self.request.redirect_url

        self.request.reset_headers()
        self.request.header('Location', location)
        self.app.bind('Response', 'redirecting ...')

        return self.data()

    def to_bytes(self):
        """Converts the data to bytes so the WSGI server can handle it.

        Returns:
            bytes -- The converted response to bytes.
        """
        return bytes(self.converted_data(), 'utf-8')

class Responsable:

    def get_response(self):
        raise NotImplementedError("This class does not implement a 'get_response()' method")



from .Manager import Manager
from .AuthManager import AuthManager
from .BroadcastManager import BroadcastManager
from .CacheManager import CacheManager
from .MailManager import MailManager
from .QueueManager import QueueManager
from .SessionManager import SessionManager
from .StorageManager import StorageManager
from .UploadManager import UploadManager



"""Migrate Status Command."""
import os
import sys

from subprocess import check_output
from cleo import Command
from ..packages import add_venv_site_packages


class MigrateStatusCommand(Command):
    """
    Migrate status.

    migrate:status
    """

    def handle(self):
        sys.path.append(os.getcwd())
        try:
            add_venv_site_packages()
        except ImportError:
            self.comment(
                'This command must be ran inside of the root of a Masonite project directory')

        from wsgi import container

        migration_directory = ['databases/migrations']
        for key, value in container.providers.items():
            if isinstance(key, str) and 'MigrationDirectory' in key:
                migration_directory.append(value)

        for directory in migration_directory:
            self.line('')
            if len(migration_directory) > 1:
                self.info('Migrate Status: {}'.format(directory))
            try:
                output = bytes(check_output(
                    ['orator', 'migrate:status', '-c',
                        'config/database.py', '-p', directory]
                )).decode('utf-8')

                self.line(
                    output.replace('Yes', '<info>Yes</info>')
                    .replace('No', '<comment>No</comment>'))
            except Exception:
                pass



"""A LoginController Module."""

from masonite.auth import Auth
from masonite.request import Request
from masonite.view import View
from masonite.helpers import config


class LoginController:
    """Login Form Controller."""

    def __init__(self):
        """LoginController Constructor."""
        pass

    def show(self, request: Request, view: View, auth: Auth):
        """Show the login page.

        Arguments:
            request {masonite.request.Request} -- The Masonite request class.
            view {masonite.view.View} -- The Masonite view class.
            auth {masonite.auth.auth} -- The Masonite auth class.

        Returns:
            masonite.view.View -- Returns the Masonite view class.
        """
        if request.user():
            return request.redirect('/home')
        return view.render('auth/login', {'app': config('application'), 'Auth': auth})

    def store(self, request: Request, auth: Auth):
        """Login the user.

        Arguments:
            request {masonite.request.Request} -- The Masonite request class.
            auth {masonite.auth.auth} -- The Masonite auth class.

        Returns:
            masonite.request.Request -- The Masonite request class.
        """
        if auth.login(request.input('email'), request.input('password')):
            return request.redirect('/home')

        return request.redirect('/login')

    def logout(self, request: Request, auth: Auth):
        """Log out the user.

        Arguments:
            request {masonite.request.Request} -- The Masonite request class.
            auth {masonite.auth.auth} -- The Masonite auth class.

        Returns:
            masonite.request.Request -- The Masonite request class.
        """
        auth.logout()
        return request.redirect('/login')



"""Auth Manager Module."""

from .Manager import Manager


class AuthManager(Manager):
    """Manages all auth drivers.

    Arguments:
        Manager {from .managers.Manager} -- The base Manager class.
    """

    config = 'auth'
    driver_prefix = 'Auth'


class Auth:
    """Dummy class that will be used to swap out the manager in the container."""

    pass



"""Cryptographic Signing Module."""
import binascii

from cryptography.fernet import Fernet

from ..exceptions import InvalidSecretKey


class Sign:
    """Cryptographic signing class."""

    def __init__(self, key=None):
        """Sign constructor.

        Keyword Arguments:
            key {string} -- The secret key to use. If nothing is passed it then it will use
                            the secret key from the config file. (default: {None})

        Raises:
            InvalidSecretKey -- Thrown if the secret key does not exist.
        """
        if key:
            self.key = key
        else:
            from config import application
            self.key = application.KEY

        if not self.key:
            raise InvalidSecretKey(
                "The encryption key passed in is: None. Be sure there is a secret key present in your .env file or your config/application.py file.")

        self.encryption = None

    def sign(self, value):
        """Sign a value using the secret key.

        Arguments:
            value {string} -- The value to be encrypted.

        Returns:
            string -- Returns the encrypted value.

        Raises:
            InvalidSecretKey -- Thrown if the secret key has incorrect padding.
        """
        try:
            f = Fernet(self.key)
        except (binascii.Error, ValueError):
            raise InvalidSecretKey(
                "You have passed an invalid secret key of: {}. Make sure you have correctly added your secret key.".format(self.key))

        self.encryption = f.encrypt(bytes(str(value), 'utf-8'))
        return self.encryption.decode('utf-8')

    def unsign(self, value=None):
        """Unsign the value using the secret key.

        Keyword Arguments:
            value {string} -- The value to be unencrypted. (default: {None})

        Returns:
            string -- Returns the unencrypted value.
        """
        f = Fernet(self.key)

        if not value:
            return f.decrypt(self.encryption).decode('utf-8')
        return f.decrypt(bytes(str(value), 'utf-8')).decode('utf-8')



"""Exception Handler Module.

A module for controlling exceptions handling when an error occurs doing executing
code in a Masonite application. These errors could are thrown during runtime.
"""

import inspect
import os
import platform
import sys
import traceback

from .app import App
from .exceptions import DumpException
from .request import Request
from .response import Response
from .view import View
from .helpers import config
from .listeners import BaseExceptionListener

package_directory = os.path.dirname(os.path.realpath(__file__))


class ExceptionHandler:
    """Class for handling exceptions thrown during runtime."""

    def __init__(self, app):
        """ExceptionHandler constructor. Also responsible for loading static files into the container.

        Arguments:
            app {masonite.app.App} -- Container object
        """
        self._app = app
        self.response = self._app.make(Response)

        self._register_static_files()

    def _register_static_files(self):
        """Register static files into the container."""
        storage = config('storage')
        if storage:
            storage.STATICFILES.update(
                {
                    os.path.join(package_directory, 'snippets/exceptions'):
                    '_exceptions/'
                }
            )

    def load_exception(self, exception):
        """Load the exception thrown into this handler.

        Arguments:
            exception {Exception} -- This is the exception object thrown at runtime.
        """
        self._exception = exception

        if self._app.has('Exception{}Handler'.format(exception.__class__.__name__)):

            return self._app.make('Exception{}Handler'.format(exception.__class__.__name__)).handle(exception)

        self.handle(exception)

    def run_listeners(self, exception, stacktraceback):
        for exception_class in self._app.collect(BaseExceptionListener):
            if '*' in exception_class.listens or exception.__class__ in exception_class.listens:
                file, line = self.get_file_and_line(stacktraceback)
                self._app.resolve(exception_class).handle(exception, file, line)

    def get_file_and_line(self, stacktraceback):
        for stack in stacktraceback[::-1]:
            if 'site-packages' not in stack[0]:
                return (stack[0], stack[1])

    def handle(self, exception):
        """Render an exception view if the DEBUG configuration is True. Else this should not return anything.

        Returns:
            None
        """
        stacktraceback = traceback.extract_tb(sys.exc_info()[2])
        self.run_listeners(exception, stacktraceback)

        request = self._app.make('Request')

        request.status(500)

        # Run Any Framework Exception Hooks
        self._app.make('HookHandler').fire('*ExceptionHook')

        # Check if DEBUG is False
        from config import application
        if not application.DEBUG:
            return

        exc_type, _, _ = sys.exc_info()
        # return a view
        if request.header('Content-Type') == 'application/json' or request.header('HTTP_ACCEPT') == 'application/json':
            stacktrace = []
            for stack in traceback.extract_tb(sys.exc_info()[2]):
                stacktrace.append("{} line {} in {}".format(stack[0], stack[1], stack[2]))

            if hasattr(exc_type, 'file'):
                stacktrace.append("{} line {} in {}".format(exc_type.file, exc_type.tb[1], exc_type.tb[2]))

            self.response.view({'error': {'exception': str(self._exception), 'status': 500, 'stacktrace': stacktrace}}, status=request.get_status())
            return

        if hasattr(exc_type, 'file'):
            last_stacktrace = [[exc_type.file, exc_type.tb[-1][1], exc_type.tb[-1][2]]]
            second_to_last_stacktrace = [[exc_type.file, exc_type.tb[-2][1], exc_type.tb[-2][2]]]
        else:
            last_stacktrace = []
            second_to_last_stacktrace = []

        self.response.view(self._app.make('View')(config('application.templates.exceptions', '/masonite/snippets/exception'),
                                               {
                                                   'exception': self._exception,
                                                   'split_exception': str(self._exception).split(' '),
                                                   'traceback': traceback,
                                                   'tb': sys.exc_info(),
                                                   'stacktrace': traceback.extract_tb(sys.exc_info()[2]) + last_stacktrace + second_to_last_stacktrace,
                                                   'second_to_last': second_to_last_stacktrace,
                                                   'app': self._app,
                                                   'providers': config('providers.providers', []),
                                                   'enumerate': enumerate,
                                                   'open': open,
                                                   'platform': platform
                                               }
        ))


class DD:

    def __init__(self, container):
        self.app = container

    def dump(self, obj):
        self.app.bind('ObjDump', obj)
        raise DumpException


class DumpHandler:

    def __init__(self, view: View, request: Request, app: App, response: Response):
        self.view = view
        self.request = request
        self.app = app
        self.response = response

    def handle(self, _):
        from config.database import Model
        self.app.make('HookHandler').fire('*ExceptionHook')

        self.response.view(self.view.render(
            '/masonite/snippets/exceptions/dump', {
                'obj': self.app.make('ObjDump'),
                'type': type,
                'list': list,
                'inspect': inspect,
                'members': inspect.getmembers(self.app.make('ObjDump'), predicate=inspect.ismethod),
                'properties': inspect.getmembers(self.app.make('ObjDump')),
                'hasattr': hasattr,
                'getattr': getattr,
                'Model': Model,
                'isinstance': isinstance,
                'show_methods': (bool, str, list, dict),
                'len': len,
            }))



"""Async Driver Method."""

import inspect
import os
from concurrent.futures import (ProcessPoolExecutor, ThreadPoolExecutor,
                                as_completed)

from ...app import App
from ...contracts import QueueContract
from ...drivers import BaseQueueDriver
from ...exceptions import QueueException
from ...helpers import HasColoredCommands, config


class QueueAsyncDriver(BaseQueueDriver, HasColoredCommands, QueueContract):
    """Queue Aysnc Driver."""

    def __init__(self, app: App):
        """Queue Async Driver.

        Arguments:
            Container {masonite.app.App} -- The application container.
        """
        self.container = app

    def _get_processor(self, mode, max_workers):
        """Set processor to use either threads or multiprocesses

        Arguments:
            mode {str} - async mode
            max_workers {int} - number of threads/processes to use
        """

        # Necessary for Python 3.4, can be removed in 3.5+
        if max_workers is None:
            # Use this number because ThreadPoolExecutor is often
            # used to overlap I/O instead of CPU work.
            max_workers = (os.cpu_count() or 1) * 5
        if max_workers <= 0:
            raise QueueException("max_workers must be greater than 0")

        # Determine Mode for Processing
        if mode == 'threading':
            processor = ThreadPoolExecutor(max_workers)
        elif mode == 'multiprocess':
            processor = ProcessPoolExecutor(max_workers)
        else:
            raise QueueException('Queue mode {} not recognized'.format(mode))
        return processor

    def push(self, *objects, args=(), kwargs={}, **options):
        """Push objects onto the async stack.

        Arguments:
            objects {*args of objects} - This can be several objects as parameters into this method.
            options {**kwargs of options} - Additional options for async driver
        """

        # Initialize Extra Options
        callback = options.get('callback', 'handle')
        mode = options.get('mode', config('queue.drivers.async.mode', 'threading'))
        workers = options.get('workers', None)

        # Set processor to either use threads or processes
        processor = self._get_processor(mode=mode, max_workers=workers)
        is_blocking = config('queue.drivers.async.blocking', False)

        ran = []
        for obj in objects:
            obj = self.container.resolve(obj) if inspect.isclass(obj) else obj
            try:
                future = processor.submit(
                    getattr(obj, callback), *args, **kwargs)
            except AttributeError:
                # Could be wanting to call only a method asyncronously
                future = processor.submit(fn=obj, *args, **kwargs)

            ran.append(future)

        if is_blocking:
            for job in as_completed(ran):
                self.info("Job Ran: {}".format(job))



import io
import json
import sys
import unittest
from contextlib import contextmanager
from urllib.parse import urlencode

from .. import env
from ..exceptions import RouteNotFoundException
from ..helpers.migrations import Migrations
from ..helpers.routes import create_matchurl, flatten_routes
from ..testsuite import generate_wsgi
from orator.orm import Factory
from ..app import App

from .MockRoute import MockRoute


class TestCase(unittest.TestCase):

    sqlite = True
    transactions = True
    refreshes_database = False
    _transaction = False
    _with_subdomains = False

    def setUp(self):
        from wsgi import container
        self.container = container

        self.acting_user = False
        self.factory = Factory()
        self.withoutExceptionHandling()
        self.withoutCsrf()
        if not self._transaction:
            self.startTransaction()
            if hasattr(self, 'setUpFactories'):
                self.setUpFactories()

        if self.sqlite and env('DB_CONNECTION') != 'sqlite':
            raise Exception("Cannot run tests without using the 'sqlite' database.")

        if not self.transactions and self.refreshes_database:
            self.refreshDatabase()

        self.route_middleware = False
        self.http_middleware = False

    def buildOwnContainer(self):
        self.container = self.create_container()
        return self

    @classmethod
    def setUpClass(cls):
        cls.staticSetUpDatabase()

    @classmethod
    def tearDownClass(cls):
        if not cls.refreshes_database and cls.transactions:
            cls.staticStopTransaction()
        else:
            cls.staticTearDownDatabase()

    def refreshDatabase(self):
        if not self.refreshes_database and self.transactions:
            self.stopTransaction()
            self.startTransaction()
            if hasattr(self, 'setUpFactories'):
                self.setUpFactories()
        else:
            self.tearDownDatabase()
            self.setUpDatabase()

    def startTransaction(self):
        from config.database import DB
        DB.begin_transaction()
        self.__class__._transaction = True

    def stopTransaction(self):
        from config.database import DB
        DB.rollback()
        self.__class__._transaction = False

    @classmethod
    def staticStopTransaction(cls):
        from config.database import DB
        DB.rollback()
        cls._transaction = False

    def make(self, model, factory, amount=50):
        self.registerFactory(model, factory)
        self.makeFactory(model, amount)

    def makeFactory(self, model, amount):
        return self.factory(model, amount).create()

    def registerFactory(self, model, callable_factory):
        self.factory.register(model, callable_factory)

    def setUpDatabase(self):
        self.tearDownDatabase()
        Migrations().run()
        if hasattr(self, 'setUpFactories'):
            self.setUpFactories()

    def tearDownDatabase(self):
        Migrations().reset()

    @staticmethod
    def staticSetUpDatabase():
        Migrations().run()

    @staticmethod
    def staticTearDownDatabase():
        Migrations().reset()

    def tearDown(self):
        if not self.transactions and self.refreshes_database:
            self.tearDownDatabase()

        if self.container.has('Request'):
            self.container.make('Request').get_and_reset_headers()

    def call(self, method, url, params, wsgi={}):
        custom_wsgi = {
            'PATH_INFO': url,
            'REQUEST_METHOD': method
        }

        custom_wsgi.update(wsgi)
        if not self._with_csrf:
            params.update({'__token': 'tok'})
            custom_wsgi.update({
                'HTTP_COOKIE': 'csrf_token=tok',
                'CONTENT_LENGTH': len(str(json.dumps(params))),
                'wsgi.input': io.BytesIO(bytes(json.dumps(params), 'utf-8')),
            })

        custom_wsgi.update({
            'QUERY_STRING': urlencode(params),
        })

        self.run_container(custom_wsgi)
        self.container.make('Request').request_variables = params
        return self.route(url, method)

    def get(self, url, params={}, wsgi={}):
        return self.call('GET', url, params, wsgi=wsgi)

    def withSubdomains(self):
        self._with_subdomains = True
        return self

    def json(self, method, url, params={}):
        return self.call(method, url, params, wsgi={
            'CONTENT_TYPE': 'application/json',
            'CONTENT_LENGTH': len(str(json.dumps(params))),
            'wsgi.input': io.BytesIO(bytes(json.dumps(params), 'utf-8')),
        })

    def post(self, url, params={}):
        return self.call('POST', url, params)

    def put(self, url, params={}):
        return self.json('PUT', url, params)

    def patch(self, url, params={}):
        return self.json('PATCH', url, params)

    def delete(self, url, params={}):
        return self.json('DELETE', url, params)

    def actingAs(self, user):
        if not user:
            raise TypeError("Cannot act as a user of type: {}".format(type(user)))
        self.acting_user = user
        return self

    def route(self, url, method=False):
        for route in self.container.make('WebRoutes'):
            matchurl = create_matchurl(url, route)
            if self.container.make('Request').has_subdomain():
                # Check if the subdomain matches the correct routes domain
                if not route.has_required_domain():
                    continue

            if matchurl.match(url) and method in route.method_type:
                return MockRoute(route, self.container)

        raise RouteNotFoundException("Could not find a route based on the url '{}'".format(url))

    def routes(self, routes=[], only=False):
        if only:
            self.container.bind('WebRoutes', flatten_routes(only))
            return

        self.container.bind('WebRoutes', flatten_routes(self.container.make('WebRoutes') + routes))

    @contextmanager
    def captureOutput(self):
        new_out, new_err = io.StringIO(), io.StringIO()
        old_out, old_err = sys.stdout, sys.stderr
        try:
            sys.stdout, sys.stderr = new_out, new_err
            yield sys.stdout
        finally:
            sys.stdout, sys.stderr = old_out, old_err

    def run_container(self, wsgi_values={}):
        wsgi = generate_wsgi()
        wsgi.update(wsgi_values)
        self.container.bind('Environ', wsgi)
        self.container.make('Request')._test_user = self.acting_user
        self.container.make('Request').load_app(self.container)
        if self._with_subdomains:
            self.container.make('Request').activate_subdomains()

        if self.route_middleware is not False:
            print('bind new middleware', self.route_middleware)
            self.container.bind('RouteMiddleware', self.route_middleware)

        if self.http_middleware is not False:
            self.container.bind('HttpMiddleware', self.http_middleware)

        try:
            for provider in self.container.make('WSGIProviders'):
                self.container.resolve(provider.boot)
        except Exception as e:
            if self._exception_handling:
                self.container.make('ExceptionHandler').load_exception(e)
            else:
                raise e

    def withExceptionHandling(self):
        self._exception_handling = True

    def withoutExceptionHandling(self):
        self._exception_handling = False

    def withCsrf(self):
        self._with_csrf = True
        return self

    def withoutCsrf(self):
        self._with_csrf = False
        return self

    def assertDatabaseHas(self, schema, value):
        from config.database import DB

        table = schema.split('.')[0]
        column = schema.split('.')[1]

        self.assertTrue(DB.table(table).where(column, value).first())

    def assertDatabaseNotHas(self, schema, value):
        from config.database import DB

        table = schema.split('.')[0]
        column = schema.split('.')[1]

        self.assertFalse(DB.table(table).where(column, value).first())

    def on_bind(self, obj, method):
        self.container.on_bind(obj, method)
        return self

    def withRouteMiddleware(self, middleware):
        self.route_middleware = middleware
        return self

    def withHttpMiddleware(self, middleware):
        self.http_middleware = middleware
        return self

    def withoutHttpMiddleware(self):
        self.http_middleware = []
        return self

    def create_container(self):
        container = App()
        from config import providers

        container.bind('WSGI', generate_wsgi())
        # container.bind('Application', application)
        container.bind('Container', container)

        # container.bind('ProvidersConfig', providers)
        container.bind('Providers', [])
        container.bind('WSGIProviders', [])

        """Bind all service providers
        Let's register everything into the Service Container. Once everything is
        in the container we can run through all the boot methods. For reasons
        some providers don't need to execute with every request and should
        only run once when the server is started. Providers will be ran
        once if the wsgi attribute on a provider is False.
        """

        for provider in providers.PROVIDERS:
            located_provider = provider()
            located_provider.load_app(container).register()
            if located_provider.wsgi:
                container.make('WSGIProviders').append(located_provider)
            else:
                container.make('Providers').append(located_provider)

        for provider in container.make('Providers'):
            container.resolve(provider.boot)

        """Get the application from the container
        Some providers may change the WSGI Server like wrapping the WSGI server
        in a Whitenoise container for an example. Let's get a WSGI instance
        from the container and pass it to the application variable. This
        will allow WSGI servers to pick it up from the command line
        """

        return container



"""Mailgun Driver Module."""
import requests

from ...contracts.MailContract import MailContract
from ...drivers import BaseMailDriver


class MailMailgunDriver(BaseMailDriver, MailContract):
    """Mailgun driver."""

    def send(self, message=None):
        """Send the message through the Mailgun service.

        Keyword Arguments:
            message {string} -- The message to be sent to Mailgun. (default: {None})

        Returns:
            requests.post -- Returns the response as a requests object.
        """

        if self._queue:
            from wsgi import container
            from .. import Queue
            return container.make(Queue).push(self._send_mail, args=(message,))

        return self._send_mail(message)

    def _send_mail(self, message):
        """Wrapper around sending mail so it can also be used with queues.

        Arguments:
            message {string|None} -- The message to be sent passed in from the send method.

        Returns:
            requests.post
        """
        if not message:
            message = self.message_body

        domain = self.config.DRIVERS['mailgun']['domain']
        secret = self.config.DRIVERS['mailgun']['secret']

        return requests.post(
            "https://api.mailgun.net/v3/{0}/messages".format(domain),
            auth=("api", secret),
            data={
                "from": "{0} <mailgun@{1}>".format(self.config.FROM['name'], domain),
                "to": [self.to_address, "{0}".format(self.config.FROM['address'])],
                "subject": self.message_subject,
                "h:Reply-To": self.message_reply_to,
                "html": message})



"""Manager Module."""

import inspect

from ..exceptions import (DriverNotFound,
                                 MissingContainerBindingNotFound,
                                 UnacceptableDriverType)

from ..app import App
from ..helpers import config


class Manager:
    """Base Manager Class."""

    config = None
    driver_prefix = None

    def __init__(self, container: App):
        """Manager constructor.

        Keyword Arguments:
            container {masonite.app.App} -- The container class (default: {None})
        """
        self.manage_driver = None
        self.container = container

    def load_container(self, container):
        """Load the container into the class and creates the default driver.

        Arguments:
            container {masonite.app.App} -- The container class

        Returns:
            self
        """
        self.container = container
        self.create_driver()
        return self

    def driver(self, driver):
        """Create the driver specified and returns the driver instance.

        Arguments:
            driver {masonite.drivers.Driver} -- An instance of a Driver class.

        Returns:
            masonite.drivers.Driver -- Returns a driver which is an instance of the base Driver class.
        """
        self.create_driver(driver)
        return self.manage_driver.load_manager(self)

    def create_driver(self, driver=None):
        """Create the driver to be used.

        This could be used as the default driver when the manager is created or called internally on the fly
        to change to a specific driver

        Keyword Arguments:
            driver {string} -- The name of the driver to switch to (default: {None})

        Raises:
            UnacceptableDriverType -- Raised when a driver passed in is not a string or a class
            DriverNotFound -- Raised when the driver can not be found.
        """

        if driver in (None, 'default'):
            driver = config('{}.driver'.format(self.config)).capitalize()
        else:
            if isinstance(driver, str):
                driver = driver.capitalize()

        try:
            if isinstance(driver, str):
                self.manage_driver = self.container.make(
                    '{0}{1}Driver'.format(self.driver_prefix, driver)
                )
                return
            elif inspect.isclass(driver):
                self.manage_driver = self.container.resolve(driver)
                return

            raise UnacceptableDriverType(
                'String or class based driver required. {} driver recieved.'.format(driver))
        except MissingContainerBindingNotFound:
            raise DriverNotFound(
                'Could not find the {0}{1}Driver from the service container. Are you missing a service provider?'.format(self.driver_prefix, driver))



"""New Controller Command."""
from ..view import View
from ..app import App
from ..helpers.filesystem import make_directory

from cleo import Command


class ControllerCommand(Command):
    """
    Creates a controller.

    controller
        {name : Name of the controller you would like to create}
        {--r|--resource : Create a controller as a resource}
        {--e|--exact : For add the name controller without `Controller` text}
    """

    def handle(self):
        controller = self.argument('name')
        view = View(App())

        if not self.option('exact'):
            controller = controller + "Controller"

        if not make_directory('app/http/controllers/{0}.py'.format(controller)):
            return self.error('{0} Controller Exists!'.format(controller))

        with open('app/http/controllers/{0}.py'.format(controller), 'w+') as f:
            if view.exists('/masonite/snippets/scaffold/controller'):
                if self.option('resource'):
                    template = '/masonite/snippets/scaffold/controller_resource'
                else:
                    template = '/masonite/snippets/scaffold/controller'

                f.write(
                    view.render(
                        template, {'class': controller.split('/')[-1]}).rendered_template
                )

        self.info('Controller Created Successfully!')



"""Module for the ache disk driver."""

import glob
import os
import time

from ...app import App
from ...contracts import CacheContract
from ...drivers import BaseCacheDriver
from ...helpers import config


class CacheDiskDriver(CacheContract, BaseCacheDriver):
    """Class for the cache disk driver."""

    def __init__(self, app: App):
        """Cache disk driver constructor.

        Arguments:
            CacheConfig {config.cache} -- Cache configuration module.
            Application {config.application} -- Application configuration module.
        """
        self.config = config('cache')
        self.appconfig = config('application')
        self.cache_forever = None

    def store(self, key, value, extension=".txt", location=None):
        """Store content in cache file.

        Arguments:
            key {string} -- The key to store the cache file into
            value {string} -- The value you want to store in the cache

        Keyword Arguments:
            extension {string} -- the extension you want to append to the file (default: {".txt"})
            location {string} -- The path you want to store the cache into. (default: {None})

        Returns:
            string -- Returns the key
        """
        self.cache_forever = True
        if not location:
            location = self.config.DRIVERS['disk']['location']

        location += '/'
        path = os.path.join(location, key + extension)
        if not os.path.exists(path):
            self._create_directory(path)

        with open(path, 'w') as file:
            file.write(value)

        return key

    def store_for(self, key, value, cache_time, cache_type, extension=".txt", location=None):
        """Store the cache for a specific amount of time.

        Arguments:
            key {string} -- The key to store the cache file into
            value {string} -- The value you want to store in the cache
            cache_time {int|string} -- The time as a string or an integer (1, 2, 5, 100, etc)
            cache_type {string} -- The type of time to store for (minute, minutes, hours, seconds, etc)

        Keyword Arguments:
            extension {string} -- the extension you want to append to the file (default: {".txt"})
            location {string} -- The path you want to store the cache into. (default: {None})

        Raises:
            ValueError -- Thrown if an invalid cache type was caught (like houes instead of hours).

        Returns:
            string -- Returns the key
        """
        self.cache_forever = False

        cache_for_time = self.calculate_time(cache_type, cache_time)

        cache_for_time = cache_for_time + time.time()

        key = self.store(
            key + ":" + str(cache_for_time),
            value, extension, location
        )

        return key

    def get(self, key):
        """Get the data from a key in the cache."""
        if not self.is_valid(key):
            return None

        cache_path = self.config.DRIVERS['disk']['location'] + "/"
        content = ""

        if self.cache_forever:
            glob_path = cache_path + key + '*'
        else:
            glob_path = cache_path + key + ':*'

        try:
            with open(glob.glob(glob_path)[0], 'r') as file:
                content = file.read()
        except IndexError:
            pass

        return content

    def delete(self, key):
        """Delete file cache."""
        cache_path = self.config.DRIVERS['disk']['location'] + "/"
        if self.cache_forever:
            glob_path = cache_path + key + '*'
        else:
            glob_path = cache_path + key + ':*'

        for template in glob.glob(glob_path):
            os.remove(template)

    def update(self, key, value, location=None):
        """Update a specific cache by key."""
        if not location:
            location = self.config.DRIVERS['disk']['location'] + "/"

        location = os.path.join(location, key)
        cache = glob.glob(location + ':*')[0]

        with open(cache, 'w') as file:
            file.write(str(value))

        return key

    def exists(self, key):
        """Check if the cache exists."""
        cache_path = self.config.DRIVERS['disk']['location'] + "/"
        if self.cache_forever:
            glob_path = cache_path + key + '*'
        else:
            glob_path = cache_path + key + ':*'

        find_template = glob.glob(glob_path)
        if find_template:
            return True
        return False

    def is_valid(self, key):
        """Check if a valid cache."""
        cache_path = self.config.DRIVERS['disk']['location'] + "/"
        if self.cache_forever:
            glob_path = cache_path + key + '*'
        else:
            glob_path = cache_path + key + ':*'

        cache_file = glob.glob(glob_path)
        if cache_file:
            try:
                cache_timestamp = float(
                    os.path.splitext(cache_file[0])[0].split(':')[1]
                )
            except IndexError:
                if self.cache_forever:
                    return True
                else:
                    return False

            if cache_timestamp > time.time():
                return True

        self.delete(key)
        return False

    def _create_directory(self, directory):
        """Creates a new dir.

        Arguments:
            directory {string} -- name of directory to create.

        Returns:
            bool - Returns a boolean if the directory was created.
        """
        if not os.path.exists(os.path.dirname(directory)):
            # Create the path to the model if it does not exist
            os.makedirs(os.path.dirname(directory))
            return True
        return False



"""A WhiteNoiseProvider Service Provider."""

from whitenoise import WhiteNoise

from ..provider import ServiceProvider
from ..helpers import config


class WhitenoiseProvider(ServiceProvider):

    wsgi = False

    def register(self):
        pass

    def boot(self):
        """Wrap the WSGI server in a whitenoise container."""
        from config import application
        self.app.bind('WSGI', WhiteNoise(
            self.app.make('WSGI'), root=config('application.static_root'), autorefresh=application.DEBUG))

        for location, alias in self.app.make('staticfiles').items():
            self.app.make('WSGI').add_files(location, prefix=alias)



"""AuthCookieDriver Module."""

from ...contracts import AuthContract
from ...drivers import BaseDriver
from ...request import Request


class AuthCookieDriver(BaseDriver, AuthContract):

    def __init__(self, request: Request):
        """AuthCookieDriver initializer.

        Arguments:
            request {masonite.request.Request} -- The Masonite request class.
        """
        self.request = request

    def user(self, auth_model):
        """Gets the user based on this driver implementation

        Arguments:
            auth_model {orator.orm.Model} -- An Orator ORM type object.

        Returns:
            Model|bool
        """
        if self.request.get_cookie('token') and auth_model:
            return auth_model.where(
                'remember_token', self.request.get_cookie('token')
            ).first()

        return False

    def save(self, remember_token, **_):
        """Saves the cookie to some state.

        In this case the state is saving to a cookie.

        Arguments:
            remember_token {string} -- A token containing the state.

        Returns:
            bool
        """
        return self.request.cookie('token', remember_token)

    def delete(self):
        """Deletes the state depending on the implementation of this driver.

        Returns:
            bool
        """
        return self.request.delete_cookie('token')
